# -*- coding: utf-8 -*-
from copy import deepcopy

from sigmaepsilon.solid.fem.structure import Structure as SolidStructure

from ...oc.oc import OC_SIMP_COMP as OC

__all__ = ['Structure']


class Structure(SolidStructure):
    
    def minimize_volume(self, inplace=True):
        pass
    
    def optimize_topology(self, *args, maxiter=300, **kwargs):
        oc = self.optimizer(*args, maxiter=maxiter,**kwargs)
        _solution = next(oc)
        yield _solution
        solution = deepcopy(_solution)
        try:
            for _ in range(maxiter):
                _solution = next(oc)
                yield _solution
                solution = deepcopy(_solution)
        except:
            return solution
        return solution
    
    def optimizer(self, *args, r_max=None, neighbours=None, **kwargs):
        if r_max is not None and neighbours is None:
            neighbours = self.mesh.cells_around_cells(r_max, 'dict')
        return OC(self, *args, neighbours=neighbours, **kwargs)


# -*- coding: utf-8 -*-
from numba import njit, prange
import numpy as np
from numpy import ndarray

from ..cells.utils import element_dof_solution_bulk, element_dof_solution_bulk_multi

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def model_strains(dofsol1d: ndarray, gnum: ndarray, B: ndarray):
    nE, nP, NSTRE = B.shape[:3]
    esol = element_dof_solution_bulk(dofsol1d, gnum)
    res = np.zeros((nE, nP, NSTRE), dtype=dofsol1d.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            res[iE, iP, :] = B[iE, iP] @ esol[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def model_strains_multi(data: ndarray, gnum: ndarray, B: ndarray):
    nE, nP, NSTRE = B.shape[:3]
    esol = element_dof_solution_bulk_multi(data, gnum)
    nRHS = esol.shape[0]
    res = np.zeros((nRHS, nE, nP, NSTRE), dtype=data.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            for j in prange(nRHS):
                res[j, iE, iP, :] = B[iE, iP] @ esol[j, iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stresses_from_strains(C: ndarray, strains: ndarray):
    nE, nP, _ = strains.shape
    res = np.zeros_like(strains)
    for iE in prange(nE):
        for iP in prange(nP):
            res[iE, iP, :] = C[iE] @ strains[iE, iP]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def stresses_from_strains_multi(C: ndarray, strains: ndarray):
    nR, nE, nP, _ = strains.shape
    res = np.zeros_like(strains)
    for iR in prange(nR):
        for iE in prange(nE):
            for iP in prange(nP):
                res[iR, iE, iP, :] = C[iE] @ strains[iR, iE, iP]
    return res



# -*- coding: utf-8 -*-
from .fem.mesh import FemMesh, SolidMesh
from .fem.pointdata import PointData
from .fem.structure import Structure
from .fem.linemesh import LineMesh
from .fem.surfacemesh import SurfaceMesh
from .model.bernoulli.section import BeamSection, get_section


# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray, swapaxes as swap
from numba import njit, prange
from scipy.interpolate import interp1d
__cache = True


def approx_element_solution_bulk(v: ndarray, A: ndarray):
    """
    Approximates discrete solution over several elements.
    
    Parameters
    ----------
    v : numpy.ndarray
        The discrete values to interpolate as an array of shape 
        (nE, nRHS, nDOF * nNE).
        
    A : numpy.ndarray
        The interpolation matrix of shape (nP, nX, nDOF * nNE) for 
        constant metrics, or (nE, nP, nX, nDOF * nNE) for variable
        matrics.
    
    Returns
    -------
    numpy.ndarray
        An array of shape (nE, nRHS, nP, nX).
        
    """
    if len(A.shape) == 4:
        res = _approx_element_solution_bulk_vm_(v, A)
    else:
        res = _approx_element_solution_bulk_cm_(v, A)
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _approx_element_solution_bulk_cm_(v: ndarray, A: ndarray):
    """
    Approximates discrete solution over several elements.
    
    Parameters
    ----------
    v : numpy.ndarray
        The discrete values to interpolate as an array of shape 
        (nE, nRHS, nDOF * nNE).
        
    A : numpy.ndarray
        The interpolation matrix of shape (nP, nX, nDOF * nNE).
    
    Returns
    -------
    numpy.ndarray
        An array of shape (nE, nRHS, nP, nX).
        
    """
    nP, nX = A.shape[:2]
    nE, nRHS = v.shape[:2]
    res = np.zeros((nE, nRHS, nP, nX), dtype=v.dtype)
    for i in prange(nE):
        for j in prange(nP):
            for k in prange(nRHS):
                res[i, k, j, :] = A[j] @ v[i, k]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _approx_element_solution_bulk_vm_(v: ndarray, A: ndarray):
    """
    Approximates discrete solution over several elements.
    
    Parameters
    ----------
    v : numpy.ndarray
        The discrete values to interpolate as an array of shape 
        (nE, nRHS, nDOF * nNE).
        
    A : numpy.ndarray
        The interpolation matrix of shape (nE, nP, nX, nDOF * nNE).
    
    Returns
    -------
    numpy.ndarray
        An array of shape (nE, nRHS, nP, nX).
        
    """
    nE, nP, nX = A.shape[:3]
    nRHS = v.shape[1]
    res = np.zeros((nE, nRHS, nP, nX), dtype=v.dtype)
    for i in prange(nE):
        for j in prange(nP):
            for k in prange(nRHS):
                res[i, k, j, :] = A[i, j] @ v[i, k]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def calculate_external_forces_bulk(K: ndarray, dofsol: ndarray):
    """
    Returns the external nodal load vectors for several elements.
    
    Parameters
    ----------
    K : numpy.ndarray
        Stiffness matrices of several elements as a 3d numpy 
        array of shape (nE, nEVAB, nEVAB).
    
    dofsol : numpy.ndarray
        Degree of freedom solution for several elements and load cases
        as a 3d numpy array of shape (nE, nRHS, nEVAB).
    
    Note
    ----
    These are the generalized forces that act on the elements.
    
    Returns
    -------
    numpy.ndarray
        3d float array of shape (nE, nRHS, nEVAB), where nE, nRHS and
        nEVAB are the number of elemnts, load cases and element variables.
        
    """
    nE, nRHS, _  = dofsol.shape
    res = np.zeros_like(dofsol)
    for i in prange(nE):
        for j in prange(nRHS):
            res[i, j] = K[i] @ dofsol[i, j]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def calculate_internal_forces_bulk(strains: ndarray, D: ndarray):
    """
    strain (nE, nRHS, nP, nX)
    D (nE, nX, nX)
    ---
    (nE, nRHS, nP, nX)
    """
    nE, nRHS, nP, _ = strains.shape
    res = np.zeros_like(strains)
    for i in prange(nE):
        for j in prange(nRHS):
            for k in prange(nP):
                res[i, j, k] = D[i] @ strains[i, j, k]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def explode_kinetic_strains(kstrains: ndarray, nP: int):
    """
    ---
    (nE, nRHS, nP, nSTRE)
    """
    nE, nSTRE, nRHS = kstrains.shape
    res = np.zeros((nE, nRHS, nP, nSTRE), dtype=kstrains.dtype)
    for i in prange(nE):
        for j in prange(nRHS):
            for k in prange(nP):
                for m in prange(nSTRE):
                    res[i, j, k, m] = kstrains[i, m, j]
    return res


def extrapolate_gauss_data_1d(gpos: ndarray, gdata: ndarray):
    gdata = swap(gdata, 0, 2)  # (nDOFN, nE, nP) --> (nP, nE, nDOFN)
    approx = interp1d(gpos, gdata, fill_value='extrapolate', axis=0)
    def inner(*args, **kwargs):
        res = approx(*args, **kwargs)
        return swap(res, 0, 2)  # (nP, nE, nDOFN) --> (nDOFN, nE, nP)
    return inner


@njit(nogil=True, parallel=True, cache=__cache)
def element_dof_solution_bulk(dofsol1d: ndarray, gnum: ndarray):
    """
    Returns an array that contains degree of freedom solution for
    every node of every element.

    Parameters
    ---------- 
    dofsol : numpy.ndarray 
        A 2d float array of shape (nX, nRHS), where nX and nRHS are
        the number of total variables and right hand sides.

    gnum : numpy.ndarray
        2d integer array of (nE, nEVAB) of global dof numbering 
        for several elements, where nE and nEVAB is the number of
        elements and total variables per element. 

    Returns
    -------
    numpy.ndarray
        3d float array of shape (nE, nEVAB, nRHS).

    """
    nRHS = dofsol1d.shape[1]
    nE, nEVAB = gnum.shape
    res = np.zeros((nE, nEVAB, nRHS), dtype=dofsol1d.dtype)
    for i in prange(nE):
        for j in prange(nRHS):
            res[i, :, j] = dofsol1d[gnum[i, :], j]
    return res


# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numpy.linalg import inv
from numba import njit, prange

from neumann.linalg import linspace1d
from neumann.array import clip1d

__cache = True


@njit(nogil=True, cache=__cache)
def layers_of_points(points: np.ndarray, bounds: np.ndarray):
    nL = bounds.shape[0]
    bins = np.zeros((nL+1,), dtype=points.dtype)
    bins[0] = bounds[0, 0]
    bins[1:] = bounds[:, 1]
    return clip1d(np.digitize(points[:, 2], bins) - 1, 0, nL-1)


@njit(['f8[:, :](f8[:, :], i8)', 'f4[:, :](f4[:, :], i8)'],
      nogil=True, parallel=True, cache=__cache)
def points_of_layers(bounds: np.ndarray, nppl=3):
    nL = bounds.shape[0]
    res = np.zeros((nL, nppl), dtype=bounds.dtype)
    for iL in prange(nL):
        res[iL] = linspace1d(bounds[iL, 0], bounds[iL, 1], nppl)
    return res


@njit(nogil=True, cache=__cache)
def glob_to_loc_layer(point: np.ndarray, bounds: np.ndarray):
    return (point[2] - bounds[0]) / (bounds[1] - bounds[0])


@njit(nogil=True, parallel=True, cache=__cache)
def iso_mindlin_plate_bulk(C: ndarray, t: ndarray, k=5/6):
    res = np.zeros((C.shape[0], 5, 5))
    for i in prange(res.shape[0]):
        res[i, :3, :3] = C[i, :3, :3] * (t[i]**3 / 12)
        res[i, 3:, 3:] = C[i, 3:, 3:] * t[i]
    res[:, 3:, 3:] *= k
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def iso_membrane_bulk(C: ndarray, t: ndarray):
    res = np.zeros((C.shape[0], 3, 3))
    for i in prange(res.shape[0]):
        res[i, :, :] = C[i, :, :] * t[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def iso_mindlin_shell_bulk(C: ndarray, t: ndarray, k=5/6):
    res = np.zeros((C.shape[0], 8, 8))
    res[:, :3, :3] = iso_membrane_bulk(C, t)
    res[:, 3:, 3:] = iso_mindlin_plate_bulk(C, t, k)
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def rotation_matrices_126(angles: np.ndarray):
    """
    Returns transformation matrmatrixices T_126 for each angle.
    Angles are expected in radians.
    """
    nL = len(angles)
    T_126 = np.zeros((nL, 3, 3), dtype=angles.dtype)
    for iL in prange(nL):
        a = angles[iL] * np.pi / 180
        T_126[iL, 0, 0] = np.cos(a)**2
        T_126[iL, 0, 1] = np.sin(a)**2
        T_126[iL, 0, 2] = -np.sin(2 * a)
        T_126[iL, 1, 0] = T_126[iL, 0, 1]
        T_126[iL, 1, 1] = T_126[iL, 0, 0]
        T_126[iL, 1, 2] = -T_126[iL, 0, 2]
        T_126[iL, 2, 0] = np.cos(a) * np.sin(a)
        T_126[iL, 2, 1] = -T_126[iL, 2, 0]
        T_126[iL, 2, 2] = np.cos(a)**2 - np.sin(a)**2
    return T_126


@njit(nogil=True, parallel=True, cache=__cache)
def rotation_matrices_45(angles: np.ndarray):
    """
    Returns transformation matrix T_45 for each angle.
    Angles are expected in radians.
    """
    nL = len(angles)
    T_45 = np.zeros((nL, 2, 2), dtype=angles.dtype)
    for iL in prange(nL):
        a = angles[iL] * np.pi / 180
        T_45[iL, 0, 0] = np.cos(a)
        T_45[iL, 0, 1] = -np.sin(a)
        T_45[iL, 1, 1] = T_45[iL, 0, 0]
        T_45[iL, 1, 0] = -T_45[iL, 0, 1]
    return T_45


@njit(nogil=True, parallel=True, cache=__cache)
def material_stiffness_matrices_126(C_126: np.ndarray, angles: np.ndarray):
    """
    Returns the components of the material stiffness matrix C_126
    in the global system.
    """
    nL = len(C_126)
    C_126_g = np.zeros_like(C_126)
    R_126 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 2]], dtype=C_126_g.dtype)
    R_126_inv = inv(R_126)
    T_126 = rotation_matrices_126(angles)
    for iL in prange(nL):
        C_126_g[iL] = T_126[iL] @ C_126[iL] @ R_126_inv @ T_126[iL].T @ R_126
    return C_126_g


@njit(nogil=True, parallel=True, cache=__cache)
def material_stiffness_matrices_45(C_45: np.ndarray, angles: np.ndarray):
    """
    Returns the components of the material stiffness matrix C_45
    in the global system.
    """
    nL = len(C_45)
    C_45_g = np.zeros_like(C_45)
    R_45 = np.array([[2, 0], [0, 2]], dtype=C_45_g.dtype)
    R_45_inv = inv(R_45)
    T_126, T_45 = rotation_matrices_45(angles)
    for iL in prange(nL):
        C_45_g[iL] = T_45[iL] @ C_45[iL] @ R_45_inv @ T_45[iL].T @ R_45
    return C_45_g



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from typing import Union, Callable, Iterable

from neumann import squeeze
from neumann.array import atleast1d, atleastnd, ascont
from neumann.utils import to_range_1d

from .utils.bernoulli import (
    shape_function_matrix_bulk, body_load_vector_bulk,
    global_shape_function_derivatives_bulk as gdshpB,
    lumped_mass_matrices_direct as dlump
)

from .metaelem import MetaFiniteElement
from ..model.beam import BernoulliBeam, calculate_shear_forces


__all__ = ['BernoulliBase']


class MetaBernoulli(MetaFiniteElement):
    """
    Python metaclass for safe inheritance. Throws a TypeError
    if a method tries to shadow a definition in any of the base
    classes.
    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace, *args, **kwargs)

        if hasattr(cls, 'shpfnc'):
            if cls.shpfnc is None and isinstance(cls.NNODE, int):
                # generate functions
                N = cls.NNODE

    @classmethod
    def generate_functions(cls, N):
        pass


class ABCBernoulli(metaclass=MetaBernoulli):
    """
    Helper class.
    """
    __slots__ = ()
    dofs = ()
    dofmap = ()


class BernoulliBase(BernoulliBeam):
    """
    Skeleton class for 1d finite elements, whose bending behaviour
    is governed by the Bernoulli theory. The implementations covered
    by this class are independent of the number of nodes of the cells.
    Specification of an actual finite element consists of specifying class
    level attributes only.

    See Also
    --------
    :class:`sigmaepsilon.solid.fem.cells.bernoulli2.Bernoulli2`
    :class:`sigmaepsilon.solid.fem.cells.bernoulli3.Bernoulli3`

    Note
    ----
    Among normal circumstances, you do not have to interact with this class 
    directly. Only use it if you know what you are doing and understand the 
    meaning of the inputs precisely.

    """

    qrule: str = None
    quadrature: dict = None
    shpfnc: Callable = None
    dshpfnc: Callable = None

    def shape_function_values(self, pcoords: Union[float, Iterable[float]], *args,
                              rng: Iterable = None, lengths: Iterable[float] = None, 
                              **kwargs) -> ndarray:
        """
        Evaluates the shape functions at the points specified by 'pcoords'.

        Parameters
        ----------
        pcoords : float or Iterable[float]
            Locations of the evaluation points.

        rng : Iterable, Optional
            The range in which the locations ought to be understood, 
            typically [0, 1] or [-1, 1]. Default is [0, 1].
            
        lengths : Iterable, Optional
            The lengths of the beams in the block. Default is None.

        Returns
        -------
        numpy.ndarray
            The returned array has a shape of (nP, nNE=2, nDOF=6), where nP, 
            nNE and nDOF stand for the number of evaluation points, nodes 
            per element and number of degrees of freedom, respectively.

        """
        rng = np.array([-1, 1]) if rng is None else np.array(rng)
        pcoords = atleast1d(np.array(pcoords))
        pcoords = to_range_1d(pcoords, source=rng, target=[-1, 1])
        lengths = self.lengths() if lengths is None else lengths
        return self.__class__.shpfnc(pcoords, lengths).astype(float)

    def shape_function_derivatives(self, pcoords:Union[float, Iterable[float]]=None, *args, 
                                   rng:Iterable=None, jac: ndarray = None, dshp: ndarray = None,
                                   lengths: Iterable[float] = None, **kwargs) -> ndarray:
        """
        Evaluates the shape function derivatives (up to third) at the points specified 
        by 'pcoords'. The function either returns the derivatives on the master or the actual 
        element, depending on the inputs.

        Valid combination of inputs are:

            * 'pcoords' and optionally 'jac' : this can be used to calculate the derivatives
              in both the local ('jac' is provided) and the parametric frame ('jac' is not provided).

            * 'dshp' and 'jac' : this combination can only be used to return the derivatives
              wrt. to any frame.

        Parameters
        ----------
        pcoords : float or Iterable, Optional
            Locations of the evaluation points. Default is None.

        rng : Iterable, Optional
            The range in which the locations ought to be understood, 
            typically [0, 1] or [-1, 1]. Only if 'pcoords' is provided.
            Default is [0, 1].
            
        lengths : Iterable, Optional
            The lengths of the beams in the block. Default is None.

        jac : Iterable, Optional
            The jacobian matrix as a float array of shape (nE, nP, 1, 1), evaluated for 
            each point in each cell. Default is None.

        dshp : Iterable, Optional
            The shape function derivatives of the master element as a float array of 
            shape (nP, nNE, nDOF=6, 3), evaluated at a 'nP' number of points. 
            Default is None.

        Returns
        -------
        numpy.ndarray
            The returned array has a shape of (nP, nNE=2, nDOF=6, 3), where nP, nNE and 
            nDOF stand for the number of evaluation points, nodes per element and number of 
            degrees of freedom, respectively. Number 3 refers to first, second and third 
            derivatives.

        Notes
        -----
        The returned array is always 4 dimensional, even if there is only one 
        evaluation point.

        """
        if pcoords is not None:
            lengths = self.lengths() if lengths is None else lengths
            # calculate derivatives wrt. the parametric coordinates in the range [-1, 1]
            pcoords = atleast1d(np.array(pcoords))
            rng = np.array([-1, 1]) if rng is None else np.array(rng)
            pcoords = to_range_1d(pcoords, source=rng, target=[-1, 1])
            dshp = self.__class__.dshpfnc(pcoords, lengths)
            if jac is None:
                # return derivatives wrt. the parametric frame
                return dshp.astype(float) 
            else:
                # return derivatives wrt. the local frame
                gdshpB(dshp, jac).astype(float)
        elif dshp is not None and jac is not None:
            # return derivatives wrt. the local frame
            return gdshpB(dshp, jac).astype(float)

    def shape_function_matrix(self, pcoords:Union[float, Iterable[float]]=None, *args, 
                              rng:Iterable=None, lengths: Iterable[float] = None, 
                              **kwargs) -> ndarray:
        """
        Evaluates the shape function matrix at the points specified by 'pcoords'.

        Parameters
        ----------
        pcoords : float or Iterable
            Locations of the evaluation points.

        rng : Iterable, Optional
            The range in which the locations ought to be understood, 
            typically [0, 1] or [-1, 1]. Default is [0, 1].

        Other Parameters
        ----------------
        These parameters are for advanced users and can be omitted.
        They are here only to avoid repeated evaulation of common quantities.

        lengths : Iterable, Optional
            The lengths of the beams in the block. Default is None.

        Returns
        -------
        numpy.ndarray
            The returned array has a shape of (nE, nP, nDOF, nDOF * nNE), where nE, nP, 
            nNE and nDOF stand for the number of elements, evaluation points, nodes 
            per element and number of degrees of freedom respectively.

        Notes
        -----
        The returned array is always 4 dimensional, even if there is only one 
        evaluation point.

        """
        pcoords = atleast1d(np.array(pcoords))
        rng = np.array([-1., 1.]) if rng is None else np.array(rng)
        pcoords = to_range_1d(pcoords, source=rng, target=[-1, 1])
        lengths = self.lengths() if lengths is None else lengths
        dshp = self.__class__.dshpfnc(pcoords, lengths)
        ecoords = self.local_coordinates()
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)
        shp = self.shape_function_values(pcoords, rng=rng)
        gdshp = self.shape_function_derivatives(jac=jac, dshp=dshp) 
        return shape_function_matrix_bulk(shp, gdshp).astype(float)

    def integrate_body_loads(self, values: ndarray) -> ndarray:
        """
        Returns nodal representation of body loads.

        Parameters
        ----------
        values : numpy.ndarray
            2d or 3d numpy float array of material densities of shape
            (nE, nNE * nDOF, nRHS) or (nE, nNE * nDOF), where nE, nNE, nDOF and nRHS 
            stand for the number of elements, nodes per element, number of degrees 
            of freedom and number of load cases respectively.

        Returns
        -------
        numpy.ndarray
            An array of shape (nE, nNE * 6, nRHS).

        Notes
        -----
        The returned array is always 3 dimensional, even if there is only one 
        load case.

        See Also
        --------
        :func:`~body_load_vector_bulk`

        """
        values = atleastnd(values, 3, back=True)
        # (nE, nNE * nDOF, nRHS) -> (nE, nRHS, nNE * nDOF)
        values = np.swapaxes(values, 1, 2)
        values = ascont(values)
        qpos, qweights = self.quadrature['full']
        rng = np.array([-1., 1.])
        shp = self.shape_function_values(qpos, rng=rng)  
        # (nP, nNE=2, nDOF=6)
        dshp = self.shape_function_derivatives(qpos, rng=rng)  
        # (nP, nNE=2, nDOF=6, 3)
        ecoords = self.local_coordinates()
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)  
        # (nE, nP, 1, 1)
        djac = self.jacobian(jac=jac)  # (nE, nG)
        gdshp = self.shape_function_derivatives(jac=jac, dshp=dshp)  
        # (nE, nP, nNE=2, nDOF=6, 3)
        return body_load_vector_bulk(values, shp, gdshp, djac, qweights).astype(float)

    def _postproc_local_internal_forces_(self, values: np.ndarray, *args,
                                         rng, points, cells, dofsol, **kwargs):
        """
        Documentation is at the base element at '...solid\\fem\\elem.py'
        values (nE, nP, 4, nRHS)
        dofsol (nE, nEVAB, nRHS)
        points (nP,)
        (nE, nP, 6, nRHS) out
        """
        ecoords = self.local_coordinates()[cells]
        dshp = self.shape_function_derivatives(points, rng=rng)[cells]
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)  
        # (nE, nP, 1, 1)
        gdshp = self.shape_function_derivatives(jac=jac, dshp=dshp)
        nE, _, nRHS = dofsol.shape
        nNE, nDOF = self.__class__.NNODE, self.__class__.NDOFN
        dofsol = dofsol.reshape(nE, nNE, nDOF, nRHS)
        D = self.model_stiffness_matrix()[cells]
        values = calculate_shear_forces(dofsol, values, D, gdshp)
        return values  # (nE, nP, 6, nRHS)

    @squeeze(True)
    def lumped_mass_matrix(self, *args, lumping: str = 'direct', alpha: float = 1/50,
                           frmt: str = 'full', **kwargs):
        """
        Returns the lumped mass matrix of the block.

        Parameters
        ----------
        alpha : float, Optional
            A nonnegative parameter, typically between 0 and 1/50 (see notes).
            Default is 1/20.

        lumping : str, Optional
            Controls lumping. Currently only direct lumping is available.
            Default is 'direct'.

        frmt : str, Optional
            Possible values are 'full' or 'diag'. If 'diag', only the diagonal
            entries are returned, if 'full' a full matrix is returned.
            Default is 'full'.

        """
        dbkey = self.__class__._attr_map_['M']
        if lumping == 'direct':
            dens = self.db.density
            try:
                areas = self.areas()
            except Exception:
                areas = np.ones_like(dens)
            lengths = self.lengths()
            topo = self.topology()
            ediags = dlump(dens, lengths, areas, topo, alpha)
            if frmt == 'full':
                N = ediags.shape[-1]
                M = np.zeros((ediags.shape[0], N, N))
                inds = np.arange(N)
                M[:, inds, inds] = ediags
                self.db[dbkey] = M
                return M
            elif frmt == 'diag':
                self.db[dbkey] = ediags
                return ediags
        raise RuntimeError("Lumping mode not recognized :(")



# -*- coding: utf-8 -*-
from numba import njit, prange
import numpy as np
from numpy import ndarray

from .solid import Solid

from ...material.utils import HMH_3d

__cache = True


__all__ = ['Solid3d']


_NSTRE_ = 6
_NDOFN_ = 3


@njit(nogil=True, parallel=True, cache=__cache)
def strain_displacement_matrix(dshp: ndarray, jac: ndarray):
    nE = jac.shape[0]
    nP, nN = dshp.shape[:2]
    nTOTV = nN * _NDOFN_
    B = np.zeros((nE, nP, _NSTRE_, nTOTV), dtype=dshp.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            gdshp = dshp[iP] @ np.linalg.inv(jac[iE, iP])
            for i in prange(nN):
                B[iE, iP, 0, 0 + i*_NDOFN_] = gdshp[i, 0]
                B[iE, iP, 1, 1 + i*_NDOFN_] = gdshp[i, 1]
                B[iE, iP, 2, 2 + i*_NDOFN_] = gdshp[i, 2]
                B[iE, iP, 3, 1 + i*_NDOFN_] = gdshp[i, 2]
                B[iE, iP, 3, 2 + i*_NDOFN_] = gdshp[i, 1]
                B[iE, iP, 4, 0 + i*_NDOFN_] = gdshp[i, 2]
                B[iE, iP, 4, 2 + i*_NDOFN_] = gdshp[i, 0]
                B[iE, iP, 5, 0 + i*_NDOFN_] = gdshp[i, 1]
                B[iE, iP, 5, 1 + i*_NDOFN_] = gdshp[i, 0]
    return B


@njit(nogil=True, parallel=True, cache=__cache)
def HMH(estrs: np.ndarray):
    nE, nP = estrs.shape[:2]
    res = np.zeros((nE, nP), dtype=estrs.dtype)
    for iE in prange(nE):
        for jNE in prange(nP):
            res[iE, jNE] = HMH_3d(estrs[iE, jNE])
    return res


class Solid3d(Solid):
    """
    A model for 3d solids.
    """
    
    dofs = ('UX', 'UY', 'UZ')
    
    NDOFN = 3
    NSTRE = 6

    def strain_displacement_matrix(self, pcoords=None, *args, dshp=None, 
                                   jac=None, **kwargs) -> ndarray:
        if dshp is None and pcoords is not None:
            dshp = self.shape_function_derivatives(pcoords)
        if jac is None:
            jac = self.jacobian_matrix(dshp=dshp)
        return strain_displacement_matrix(dshp, jac)

    @classmethod
    def HMH(cls, stresses:ndarray, *args, **kwargs) -> ndarray:
        return HMH(stresses)



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray

from dewloosh.core import classproperty
from neumann.array import atleastnd
from neumann.array import atleastnd
from neumann.logical import is1dfloatarray, isboolarray
from polymesh.celldata import CellData as MeshCellData


class CellData(MeshCellData): 
    """
    A subclass of :class:`polymesh.celldata.CellData`to handle data related 
    to the cells of a finite element mesh. This class does not need to be used 
    directly, but rather serves as a base class for finite element classes.
    
    Parameters
    ----------
    activity : numpy.ndarray, Optional
        1d boolean array describing the activity of the elements.
        
    density : numpy.ndarray, Optional
        1d float array describing the density as mass per unit volume if areas
        (or thickness) are provided, mass per unit length (or area) othervise. 
        See the Notes. Default is None.
        
    loads : numpy.ndarray, Optional
        3d (for a single load case) or 4d (for multiple load cases) float 
        array of body loads for each load component of each node of each cell.
        Default is None.
        
    strain_loads : numpy.ndarray, Optional
        2d float array of body strain loads for each cell and strain 
        component. Default is None.
        
    t or thickness : numpy.ndarray, Optional
        1d float array of thicknesses. Only for 2d cells.
        Default is None.
            
    fixity : numpy.ndarray, Optional
        3d boolean array of element fixity. Default is None.
        
    areas : numpy.ndarray, Optional
        1d float array of cross sectional areas. Only for 1d cells.
        Default is None.
        
    fields : dict, Optional
        Every value of this dictionary is added to the dataset. 
        Default is `None`.
        
    **kwargs : dict, Optional
        For every key and value pair where the value is a numpy array
        with a matching shape (has entries for all points), the key
        is considered as a field and the value is added to the database.
        
    Notes
    -----
    For 1d and 2d cells, it is the user's responsibility to ensure that the 
    input data makes sense. For example, 'areas' can be omitted, but then 
    'density' must be 'mass per unit length'. The same way, 'thickness' can 
    be omitted for 2d cells if 'density' is 'mass per unit area'. Otherwise 
    'density' is expected as 'mass per unit volume'.
    
    See also
    --------
    :class:`polymesh.celldata.CellData`
    :class:`sigmaepsilon.solid.fem.pointdata.PointData`
        
    Example
    -------
    All cell types are subclasses of :class:`CellData`
    
    >>> from sigmaepsilon.solid.fem.cells import B2 as Beam
    >>> import numpy as np
    >>> random_topo = np.zeros((10, 2), dtype=int)  # random erroneous topology
    >>> random_data = np.random.rand(10, 3)
    >>> celldata = Beam(topo=random_topo, random_field = random_data)
    
    Then, the attached data is available as either
    
    >>> celldata['random_field']
    
    or 
    
    >>> celldata.random_field
        
    """

    _attr_map_ = {
        'body-loads': '_body-loads_',
        'strain-loads': '_strain-loads_',
        'density': '_density_',
        'activity' : '_activity_',
        'fixity' : '_fixity_',
        'areas' : '_areas_',
        'K': '_K_',  # stiffness matrix
        'M': '_M_',  # mass matrix
        'B': '_B_',  # strain displacement matrix
        'f' : '_f_' # nodal load vector
    }
    
    def __init__(self, *args, model:ndarray=None, activity:ndarray=None, 
                 density:ndarray=None, loads:ndarray=None, fields:dict=None,
                 strain_loads:ndarray=None, t:ndarray=None, thickness:ndarray=None, 
                 fixity:ndarray=None, areas:ndarray=None, **kwargs):
        amap = self.__class__._attr_map_

        t = t if thickness is None else thickness
        if fields is not None:
            activity = fields.pop(amap['activity'], activity)
            loads = fields.pop(amap['body-loads'], loads)
            strain_loads = fields.pop(amap['strain-loads'], strain_loads)
            density = fields.pop(amap['density'], density)
            t = fields.pop(amap['t'], t)

        super().__init__(*args, fields=fields, activity=activity, **kwargs)

        if self.db is not None:
            topo = self.topology()
            nE, nNE = topo.shape
            NDOFN = self.__class__.NDOFN

            if activity is None:
                activity = np.ones(nE, dtype=bool)
            else:
                assert isboolarray(activity) and len(activity.shape) == 1, \
                    "'activity' must be a 1d boolean numpy array!"
            self.activity = activity
            
            # fixity
            if isinstance(fixity, np.ndarray):
                assert isboolarray(fixity), "Fixity must be a boolean array."
                self.fixity = fixity
                
            # areas
            if isinstance(areas, np.ndarray):
                self._wrapped[self.__class__._attr_map_['areas']] = areas
            
            # densities
            if isinstance(density, np.ndarray):
                assert len(density.shape) == 1, \
                    "'densities' must be a 1d float or integer numpy array!"
                self.density = density
            else:
                if density is not None and 'density' not in fields:
                    raise TypeError("'density' must be a 1d float" + \
                                    " or integer numpy array!")

            # body loads
            if loads is not None:
                assert isinstance(loads, np.ndarray)
                if loads.shape[0] == nE and loads.shape[1] == nNE:
                    self.loads = loads
                elif loads.shape[0] == nE and loads.shape[1] == nNE * NDOFN:
                    loads = atleastnd(loads, 3, back=True)
                    self.loads = loads.reshape(
                        nE, nNE, NDOFN, loads.shape[-1])

            # strain loads
            if strain_loads is not None:
                assert isinstance(strain_loads, np.ndarray)
                assert strain_loads.shape[0] == nE
                self.strain_loads = strain_loads

            if self.__class__.NDIM == 2:
                if t is None:
                    t = np.ones(nE, dtype=float)
                else:
                    if isinstance(t, float):
                        t = np.full(nE, t)
                    else:
                        assert is1dfloatarray(t), \
                            "'t' must be a 1d numpy array or a float!"
                self.db[amap['t']] = t

        self._model = model
        
    @classproperty
    def _dbkey_stiffness_matrix_(cls) -> str:
        return cls._attr_map_['K']
    
    @classproperty
    def _dbkey_nodal_load_vector_(cls) -> str:
        return cls._attr_map_['f']

    @classproperty
    def _dbkey_strain_displacement_matrix_(cls) -> str:
        return cls._attr_map_['B']

    @classproperty
    def _dbkey_mass_matrix_(cls) -> str:
        return cls._attr_map_['M']
    
    @classproperty
    def _dbkey_fixity_(cls) -> str:
        return cls._attr_map_['fixity']
    
    @classproperty
    def _dbkey_density_(cls) -> str:
        return cls._attr_map_['density']
    
    @classproperty
    def _dbkey_activity_(cls) -> str:
        return cls._attr_map_['activity']
    
    @classproperty
    def _dbkey_body_loads_(cls) -> str:
        return cls._attr_map_['body-loads']
    
    @classproperty
    def _dbkey_strain_loads_(cls) -> str:
        return cls._attr_map_['strain-loads']
    
    # methods to check if a field is available
    
    @property
    def has_fixity(self):
        return self._dbkey_fixity_ in self._wrapped.fields
    
    @property
    def has_body_loads(self):
        return self._dbkey_body_loads_ in self._wrapped.fields
    
    @property
    def has_strain_loads(self):
        return self._dbkey_strain_loads_ in self._wrapped.fields
    
    @property
    def has_nodal_load_vector(self):
        return self._dbkey_nodal_load_vector_ in self._wrapped.fields
    
    # methods to access and set fields
    
    @property
    def nodes(self) -> ndarray:
        """Returns the topology of the cells."""
        return self._wrapped[self._dbkey_nodes_].to_numpy()

    @nodes.setter
    def nodes(self, value: ndarray):
        assert isinstance(value, ndarray)
        self._wrapped[self._dbkey_nodes_] = value
    
    @property
    def strain_loads(self) -> ndarray:
        """Returns strain loads."""
        return self._wrapped[self._dbkey_strain_loads_].to_numpy()

    @strain_loads.setter
    def strain_loads(self, value: ndarray):
        assert isinstance(value, ndarray)
        self._wrapped[self._dbkey_strain_loads_] = value    
    
    @property
    def loads(self) -> ndarray:
        """Returns body loads."""
        return self._wrapped[self._dbkey_body_loads_].to_numpy()

    @loads.setter
    def loads(self, value: ndarray):
        assert isinstance(value, ndarray)
        self._wrapped[self._dbkey_body_loads_] = value
                
    @property
    def density(self) -> ndarray:
        """Returns densities."""
        return self._wrapped[self._dbkey_density_].to_numpy()

    @density.setter
    def density(self, value: ndarray):
        assert isinstance(value, ndarray)
        self._wrapped[self._dbkey_density_] = value
        
    @property
    def activity(self) -> ndarray:
        """Returns activity of the cells."""
        return self._wrapped[self._dbkey_activity_].to_numpy()

    @activity.setter
    def activity(self, value: ndarray):
        """Sets the activity of the cells."""
        assert isinstance(value, ndarray)
        self._wrapped[self._dbkey_activity_] = value
    
    @property
    def fixity(self) -> ndarray:
        """Returns the fixity of the cells."""
        return self._wrapped[self._dbkey_fixity_].to_numpy()

    @fixity.setter
    def fixity(self, value: ndarray):
        """
        Sets the fixity of the cells.
        """
        assert isinstance(value, ndarray)
        self._wrapped[self._dbkey_fixity_] = value



# -*- coding: utf-8 -*-

# penalty parameter to enforce essential boundary conditions
# using a Courant penalty function
DEFAULT_DIRICHLET_PENALTY = 1E12

# a ratio of the penalty factor applied for the stiffness and
# mass matrices for the bipenalty method
DEFAULT_MASS_PENALTY_RATIO = 1E5



# -*- coding: utf-8 -*-
from neumann.numint import GaussPoints as Gauss
from polymesh.cells import QuadraticLine as Line

from .bernoulli import BernoulliBase as Bernoulli
from .gen.b3 import (shape_function_values_bulk, 
                     shape_function_derivatives_bulk)
    
from .elem import FiniteElement
from .metaelem import ABCFiniteElement as ABC
    
    
__all__ = ['Bernoulli3']


class Bernoulli3(ABC, Bernoulli, Line, FiniteElement):
    """
    Finite element class to handle 3-noded Bernoulli beams.
    """

    qrule = 'full'
    quadrature = {
        'full': Gauss(6),
        'mass' : Gauss(8)
    }
    shpfnc = shape_function_values_bulk
    dshpfnc = shape_function_derivatives_bulk



# -*- coding: utf-8 -*-
import numpy as np

from polymesh.cells import TET4 as Tetra

from ..model.solid3d import Solid3d

from .elem import FiniteElement
from .metaelem import ABCFiniteElement as ABC



class TET4(ABC, Solid3d, Tetra, FiniteElement):

    qrule = 'full'
    quadrature = {
        'full': (np.array([[1/3, 1/3, 1/3]]), np.array([1/6])),
    }



# -*- coding: utf-8 -*-
import sympy as sy

from neumann.linalg.tensor3333 import ComplianceTensor


__all__ = ['smat_sym_ortho_3d', 'cmat_sym_ortho_3d']

imap_ref_ortho_3D = ComplianceTensor.__imap__
def upper(ij): return (ij[1], ij[0]) if ij[0] >= ij[1] else (ij[0], ij[1])


def smat_sym_ortho_3d(*args, imap: dict = None, simplify=True, **subs) -> sy.Matrix:
    """
    Reurns the compliance matrix for a 3d orthotropic
    material in symbolic form.

    Parameters
    ----------
    imap : dict, optional
        Index map. Default is None.

    subs : dict, optional
        A dictionary of parameters to substitute.

    Notes
    -----
    The resulting symbolic matrix has 12 free symbols,
    but only 9 of these are independent.

    Returns
    -------
    sympy.Matrix
        6x6 symbolic compliance matrix.
        
    """
    E1, E2, E3 = sy.symbols('E1 E2 E3')
    G23, G12, G13 = sy.symbols('G23 G12 G13')
    NU12, NU21 = sy.symbols('NU12 NU21')
    NU13, NU31 = sy.symbols('NU13 NU31')
    NU23, NU32 = sy.symbols('NU23 NU32')
    S = sy.Matrix([[1/E1, -NU12/E1, -NU13/E1, 0, 0, 0],
                   [-NU21/E2, 1/E2, -NU23/E2, 0, 0, 0],
                   [-NU31/E3, -NU32/E3, 1/E3, 0, 0, 0],
                   [0, 0, 0, 1/G23, 0, 0],
                   [0, 0, 0, 0, 1/G13, 0],
                   [0, 0, 0, 0, 0, 1/G12]])
    if len(subs) > 0:
        S = S.subs([(sym, val) for sym, val in subs.items()])
    if imap is None:
        return S
    assert len(imap) == 6, "Indexmap must contain 6 indices."
    invmap = {upper(v): k for k, v in imap.items()}
    trmap = {i: invmap[imap_ref_ortho_3D[i]] for i in range(6)}
    S_new = sy.zeros(6, 6)
    for inds in trmap.items():
        S_new[inds[1], :] = swap1d(S[inds[0], :], inds)
    if simplify:
        S_new.simplify()
    return S_new


def cmat_sym_ortho_3d(*args, imap: dict = None, simplify=True, **subs):
    """
    Reurns the stiffness matrix for a 3d orthotropic
    material in symbolic form.

    Parameters
    ----------
    imap : dict, optional
        Index map. Default is None.

    subs : dict, optional
        A dictionary of parameters to substitute.

    Notes
    -----
    The resulting symbolic matrix has 12 free symbols,
    but only 9 of these are independent.

    Returns
    -------
    sympy.Matrix
        6x6 symbolic stiffness matrix 
    """
    res = smat_sym_ortho_3d(*args, imap=imap, simplify=False, **subs).inv()
    if simplify:
        res.simplify()
    return res


def swap1d(a: sy.Matrix, inds: tuple):
    a.col_swap(*inds)
    return a


if __name__ == '__main__':
    C = cmat_sym_ortho_3d()



import numpy as np
from numpy import ndarray
from numba import njit, prange

from neumann.array import repeat_diagonal_2d, ascont

__cache = True


@njit(nogil=True, cache=__cache)
def nodal_dcm(dcm: ndarray, N:int=2) -> ndarray:
    """
    Assembles the nodal direction cosine matrix of a single node.
    
    Parameters
    ----------
    dcm : numpy.ndarray
        A 3x3 direction cosine matrix.
    
    N : int, Optional
        The number of triplets that make up a nodal vector.
        Default is 2, which means 6 dofs per node.
        
    Returns
    -------
    numpy.ndarray
        A 2d numpy float array.
    
    Note
    ----
    The template 'dcm' should be the matrix you would use to transform a 
    position vector in 3d Euclidean space.
    
    """
    return repeat_diagonal_2d(dcm, N)


@njit(nogil=True, parallel=True, cache=__cache)
def nodal_dcm_bulk(dcm: ndarray, N:int=2) -> ndarray:
    """
    Assembles the nodal direction cosine matrices of several cells. 
    
    Parameters
    ----------
    dcm : numpy.ndarray
        A 3x3 direction cosine matrix for each element (4d).
    
    N : int, Optional
        The number of triplets that make up a nodal vector.
        Default is 2, which means 6 dofs per node.
        
    Returns
    -------
    numpy.ndarray
        A 3d numpy float array.
    
    Note
    ----
    Typically, 'dcm' should be the matrix you would use to transform a 
    position vector in 3d Euclidean space.
    
    """
    nE = dcm.shape[0]
    res = np.zeros((nE, 3 * N, 3 * N), dtype=dcm.dtype)
    for i in prange(N):
        _i = i * 3
        i_ = _i + 3
        res[:, _i:i_, _i:i_] = dcm
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def element_dcm(nodal_dcm: ndarray, nNE: int = 2, nDOF: int=6):
    nEVAB = nNE * nDOF
    res = np.zeros((nEVAB, nEVAB), dtype=nodal_dcm.dtype)
    for iNE in prange(nNE):
        i0, i1 = nDOF*iNE, nDOF * (iNE+1)
        res[i0: i1, i0: i1] = nodal_dcm
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def element_dcm_bulk(nodal_dcm: ndarray, nNE: int = 2, nDOF: int=6):
    nE = nodal_dcm.shape[0]
    nEVAB = nNE * nDOF
    res = np.zeros((nE, nEVAB, nEVAB), dtype=nodal_dcm.dtype)
    for iNE in prange(nNE):
        i0, i1 = nDOF*iNE, nDOF * (iNE+1)
        for iE in prange(nE):
            res[iE, i0: i1, i0: i1] = nodal_dcm[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tr_element_vectors_bulk_multi(values: ndarray, dcm: ndarray, 
                             invert:bool=False) -> ndarray:
    """
    Transforms element vectors for multiple cases.
    
    Parameters
    ----------
    values : numpy.ndarray
        The values to transform as an array of shape (nE, nRHS, nX).
    
    dcm : numpy.ndarray
        The direct cosine matrix of the transformation as an
        array of shape (nE, nX).
        
    invert : bool, Optional
        If True, the DCM matrices are transposed before transformation.
        This makes this function usable in both directions.
        Default is False.
        
    Returns
    -------
    numpy.ndarray
        An array with the same shape as 'values'.
    
    """
    res = np.zeros_like(values)
    if not invert:
        for iE in prange(res.shape[0]):
            for jRHS in prange(res.shape[1]):
                res[iE, jRHS] = dcm[iE] @ values[iE, jRHS]
    else:
        for iE in prange(res.shape[0]):
            for jRHS in prange(res.shape[1]):
                res[iE, jRHS] = dcm[iE].T @ values[iE, jRHS]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tr_element_matrices_bulk(K: ndarray, dcm: ndarray, invert:bool=False):
    """
    Transforms element stiffness matrices.
    """
    res = np.zeros_like(K)
    if not invert:
        for iE in prange(res.shape[0]):
            res[iE] = dcm[iE] @ K[iE] @ dcm[iE].T
    else:
        for iE in prange(res.shape[0]):
            res[iE] = dcm[iE].T @ K[iE] @ dcm[iE]
    return res


def tr_nodal_loads_bulk(nodal_loads: ndarray, dcm: ndarray) -> ndarray:
    """
    Transforms discrete nodal loads to the global frame.

    Parameters
    ----------
    nodal_loads : numpy.ndarray
        A 3d array of shape (nE, nEVAB, nRHS).

    Returns
    -------
    numpy.ndarray
        A numpy array of shape (nE, nEVAB, nRHS).

    """
    nodal_loads = np.swapaxes(nodal_loads, 1, 2)
    # (nE, nNE * nDOF, nRHS) -> (nE, nRHS, nNE * nDOF)
    nodal_loads = ascont(nodal_loads)
    nodal_loads = tr_element_vectors_bulk_multi(nodal_loads, dcm)
    nodal_loads = np.swapaxes(nodal_loads, 1, 2)
    # (nE, nRHS, nNE * nDOF) -> (nE, nNE * nDOF, nRHS)
    return nodal_loads


# -*- coding: utf-8 -*-
import numpy as np

from polymesh.config import __hasplotly__, __hasmatplotlib__

from .mesh import FemMesh

from .cells import Q4M, Q9M


class SurfaceMesh(FemMesh):
    """
    A data class dedicated to 2d cells. It handles sections and other line 
    related information, presets for plotting, etc.
    
    """
    
    _cell_classes_ = {
        4: Q4M,
        9: Q9M,
    }

    def __init__(self, *args, section=None, **kwargs):
        if section is not None: 
            pass
        self._section = section
        super().__init__(*args, **kwargs)
        
    @property
    def section(self):
        if self._section is not None:
            return self._section
        else:
            if self.is_root():
                return self._section
            else:
                return self.parent.section
                        


# -*- coding: utf-8 -*-
from abc import abstractmethod


class NavierProblem:
    """
    Base class for Navier problems. The sole reason of this class is to
    avoid circular referencing.
    """
    
    @abstractmethod    
    def solve(self, *args, **kwargs):
        ...    
    


# -*- coding: utf-8 -*-
import json


def json2dict(jsonpath: str):
    with open(jsonpath, "r") as jsonfile:
        return json.loads(jsonfile.read())


def dict2json(jsonpath: str, d: dict):
    with open(jsonpath, "w") as outfile:
        json.dump(d, outfile)



# -*- coding: utf-8 -*-
from scipy.sparse import coo_matrix as coo_np
from scipy.sparse import csr_matrix as csr_np
from scipy.sparse import csc_matrix as csc_np
from scipy.sparse import spmatrix
from scipy.sparse import isspmatrix as isspmatrix_np
import numpy as np
from numpy import ndarray
from numba import njit, prange
from numba.typed import Dict as nbDict
from numba import types as nbtypes

from neumann.array import minmax

__cache = True

nbint64 = nbtypes.int64
nbuint64 = nbtypes.uint64


def index_mappers(*args, **kwargs):
    if isinstance(args[0], np.ndarray):
        return index_mappers_bulk(*args, **kwargs)
    elif isspmatrix_np(args[0]):
        return index_mappers_sparse(*args, **kwargs)
    raise TypeError


def index_mappers_bulk(inds: np.ndarray, N: int = None, return_trivial: bool = False,
                       return_inverse: bool = False):
    imin, imax = minmax(inds)
    loc_to_glob = array_to_dict(np.unique(inds), 0)
    N = imax if N is None else N
    glob_to_loc = invert_mapping(loc_to_glob) if return_inverse else None
    if imin == 0 and imax == (len(loc_to_glob)-1) == (N - 1):
        if not return_trivial:
            loc_to_glob = None
            glob_to_loc = None
    if return_inverse:
        return loc_to_glob, glob_to_loc
    return loc_to_glob


def index_mappers_sparse(A: spmatrix, return_trivial: bool = False,
                         return_inverse: bool = False):
    if isinstance(A, coo_np):
        inds = A.row
    elif isinstance(A, csr_np) or isinstance(A, csc_np):
        inds = A.indices
    else:
        raise NotImplementedError
    return index_mappers_bulk(inds, A.shape[0], return_trivial, return_inverse)


def box_dof_numbering(gnum: np.ndarray, glob_to_loc: nbDict = None):
    if glob_to_loc is None:
        return gnum
    elif isinstance(glob_to_loc, nbDict):
        return box_indices2d(gnum, glob_to_loc)
    elif isinstance(glob_to_loc, ndarray):
        pass
    raise NotImplementedError


def box_spmatrix(spmat: spmatrix, glob_to_loc: nbDict = None):
    if isinstance(spmat, coo_np):
        return box_coo(spmat, glob_to_loc)
    elif isinstance(spmat, csr_np) or isinstance(spmat, csc_np):
        return box_csr(spmat, glob_to_loc)
    else:
        raise NotImplementedError('Unknown type : {}'.format(type(spmat)))


def box_coo(coo: coo_np, glob_to_loc: nbDict = None, sum_duplicates=True):
    if sum_duplicates:
        coo.sum_duplicates()
    if glob_to_loc is None:
        return coo
    elif isinstance(glob_to_loc, nbDict):
        N = len(glob_to_loc)
        coo.row = box_indices1d(coo.row, glob_to_loc)
        coo.col = box_indices1d(coo.col, glob_to_loc)
        coo._shape = (N, N)
        return coo
    elif isinstance(glob_to_loc, ndarray):
        """
        inds = np.where(glob_to_loc[coo.row] >= 0)[0]
        coo.row = coo.row[inds]
        coo.col = coo.col[inds]
        coo.data = coo.data[inds]
        pass
        """
    raise NotImplementedError


def box_csr(csr: csr_np, glob_to_loc: nbDict = None):
    """
    FIXME. This fails.
    """
    raise RuntimeWarning('FIXME. THIS FAILS!')
    if glob_to_loc is None:
        return csr
    elif isinstance(glob_to_loc, nbDict):
        N = len(glob_to_loc)
        csr.indices, csr.indptr = \
            _box_csr_data(csr.indices, csr.indptr, glob_to_loc)
        csr._shape = (N, N)
        return csr
    raise NotImplementedError


@njit(nogil=True, parallel=True, cache=__cache)
def _box_csr_data(indices: np.ndarray, indptr: np.ndarray,
                  glob_to_loc: nbDict):
    """
    FIXME. This fails.
    """
    raise RuntimeWarning('FIXME. THIS FAILS!')
    N = len(glob_to_loc)
    indices_ = np.zeros(indices.shape[0], dtype=indices.dtype)
    indptr_ = np.zeros(N + 1, dtype=indptr.dtype)
    widths_ = np.zeros(N, dtype=indptr.dtype)
    for i in prange(indices.shape[0]):
        indices_[i] = glob_to_loc[indices[i]]
    for i in prange(N):
        widths_[i] = indptr[i+1] - indptr[i]
    indptr_[1:] = widths_
    return indices_, indptr_


def box_csc(csc: csc_np, glob_to_loc: nbDict = None):
    return box_csr(csc, glob_to_loc)


def box_rhs(f: np.ndarray, loc_to_glob: nbDict = None):
    if loc_to_glob is None:
        return f
    elif isinstance(loc_to_glob, nbDict):
        return _box(f, loc_to_glob)


def unbox_lhs(u: np.ndarray, loc_to_glob: nbDict = None, N: int = None):
    if loc_to_glob is None:
        return u
    elif isinstance(loc_to_glob, nbDict):
        if N is None:
            N = np.max(list(loc_to_glob.values())) + 1
        if len(u.shape) == 1:
            u = u.reshape(len(u), 1)
        return _unbox(u, loc_to_glob, N)


@njit(nogil=True, parallel=True, cache=__cache)
def _unbox(a: np.ndarray, loc_to_glob: nbDict, N: int):
    res = np.zeros((N, a.shape[1]), dtype=a.dtype)
    for i in prange(len(loc_to_glob)):
        res[loc_to_glob[nbuint64(i)], :] = a[i, :]  
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _box(a: np.ndarray, loc_to_glob: nbDict):
    N = len(loc_to_glob)
    res = np.zeros((N, a.shape[1]), dtype=a.dtype)
    for i in prange(N):
        res[i, :] = a[loc_to_glob[nbuint64(i)], :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def box_indices1d(indices: np.ndarray, glob_to_loc: nbDict):
    """
    Returns an 1d index array, renumbered according to the 
    mapping introduced by the mapping `glob_to_loc`.
    """
    res = np.zeros_like(indices)
    for i in prange(res.shape[0]):
        res[i] = glob_to_loc[nbuint64(indices[i])]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def box_indices2d(indices: np.ndarray, glob_to_loc: nbDict):
    res = np.zeros_like(indices)
    for i in prange(res.shape[0]):
        for j in prange(res.shape[1]):
            res[i, j] = glob_to_loc[nbuint64(indices[i, j])]
    return res


@njit(nogil=True, cache=__cache)
def array_to_dict(a: np.ndarray, firstindex: int = 0):
    res = nbDict.empty(key_type=nbuint64, value_type=nbuint64)
    for i in range(len(a)):
        ui = nbuint64(firstindex + i)
        res[ui] = nbuint64(a[i])
    return res


@njit(nogil=True, cache=__cache)
def invert_mapping(imap: nbDict):
    res = nbDict.empty(key_type=nbuint64, value_type=nbuint64)
    for i in imap:
        ui = nbuint64(imap[i])
        res[ui] = nbuint64(i)
    return res



# -*- coding: utf-8 -*-
import numpy as np
from copy import deepcopy
from typing import NamedTuple, Iterable

from neumann.linalg.sparse.csr import csr_matrix as csr
from polymesh.utils import cells_around
from sigmaepsilon.solid.fem.structure import Structure

from .filter import sensitivity_filter, sensitivity_filter_csr
from .utils import (compliances_bulk, get_filter_factors,
                    get_filter_factors_csr,
                    weighted_stiffness_bulk as weighted_stiffness)


class OptRes(NamedTuple):
    """
    A tuple collecting information about an iteration block.
            
    """
    x : Iterable
    obj : float
    vol : float
    pen : float
    n : int


def maximize_stiffness(structure: Structure, *args,
                       miniter:int=50, maxiter:int=100, 
                       p_start:float=1.0, p_stop:float=3.0, p_inc:float=0.2, p_step:int=5, 
                       q:float=0.5, vfrac:float=0.6, dtol:float=0.1, r_max:float=None, 
                       penalty:float=None, nostop:bool=True, 
                       neighbours:Iterable=None, guess:Iterable=None, i_start:int=0, 
                       **kwargs) -> OptRes:
    """
    Performs topology optimization using an Optimality Criteria Method to
    maximize the stiffness of a structure, given a design space and a certain 
    amount of material to distribute.
    
    .. math::
        :nowrap:
        
        \\begin{equation}
            \\begin{array}{rrclcl}
            \\displaystyle \\min_{\\boldsymbol{\\rho}} & \\mathbf{u(\\boldsymbol{\\rho})}^T \\mathbf{f} \\\\
            \\textrm{s. t.} \\\\
            & \\mathbf{K}(\\boldsymbol{\\rho}) \\mathbf{u(\\boldsymbol{\\rho})} & = & \\mathbf{f} \\\\
            & V(\\boldsymbol{\\rho}) - \\eta V_0 & \\leq & 0 & & \\\\
            & \\rho_i \\in \\{0, 1\\} & & & & \\forall i \\in N
            \\end{array}
        \\end{equation}
    
    Parameters
    ----------
    structure : Structure
        An instance of sigmaepsilon.solid.fem.Structure.
        
    miniter : int, Optional
        The minimum number of iterations to perform. Default is 50.
        
    maxiter : int, Optional
        The maximum number of iterations to perform. Default is 100.
    
    p_start : float, Optional
        Initial value of the penalty on intermediate densities. Default is 1.
        
    p_stop : float, Optional
        Final value of the penalty on intermediate densities. Default is 3.
        
    p_inc : float, Optional
        Increment of the penalty on intermediate densities. Default is 0.2
        
    p_step : int, Optional
        The number of interations it takes to increment the penalty on intermediate
        density values. Default is 5.
        
    q : float, Optional
        Smoothing factor. Defaul is 0.5.
        
    vfrac : float, Optional
        The fraction of available volume and the volume of the virgin structure.
        Default is 0.6.
        
    dtol : float, Optional
        This controls the maximum change in the value of a design variable. 
        Default is 0.1.
        
    r_max : float, Optional
        Radius for filtering. Default is None.
        
    neighbours : float, Optional
        The neighbours of the cells for filtering. Default is None.
    
    guess : numpy.ndarray, Optional
        A guess on the solution. This parameter can be used to contiue
        a workflow. Default is None.
        
    i_start : int, Optional
        Starting index for iterations. This parameter can be used to contiue
        a workflow. Default is 0.
        
    summary : bool, Optional
        If True, a short summary about execution time and the number of iterations
        is available after execution as `structure.summary['topopt']`. Default is False.
        
    nostop : bool, Optional
        If True, iterations neglect all stopping criteria, those govern by mniniter and maxiter included.
        Default is False.
    
    Yields
    ------
    OptRes
        The results of the actual iteration.
        
    Notes
    -----
    - The function returns a generator expression.
    - This function can be used for both size and topology optimization,
      depending on the inputs.
    
    """
    do_filter = r_max is not None

    # if i_start==0:
    if structure.Solver is None:
        structure.preprocess()
    femsolver = structure.Solver.core
    assert femsolver.regular
    gnum = femsolver.gnum
    K_bulk_0 = np.copy(femsolver.K)
    vols = structure.volumes()
    centers = structure.centers()

    # initial solution to set up parameters
    dens = np.zeros_like(vols)
    dens_tmp = np.zeros_like(dens)
    dens_tmp_ = np.zeros_like(dens)
    dCdx = np.zeros_like(dens)
    comps = np.zeros_like(dens)

    def get_dof_solution():
        U = femsolver.u
        dU = len(U.shape)
        if dU == 2:
            # multiple load cases
            return U[:,  0]
        return U

    def compliance(update_stiffness=False):
        """
        Init == True means that we are in the initialization phase.
        """
        if update_stiffness:
            femsolver.update_stiffness(weighted_stiffness(K_bulk_0, dens))
        femsolver.proc()
        U = get_dof_solution()
        comps[:] = compliances_bulk(K_bulk_0, U, gnum)
        np.clip(comps, 1e-7, None, out=comps)
        return np.sum(comps)

    # ------------------ INITIAL SOLUTION ---------------
    comp = compliance()
    vol = np.sum(vols)
    vol_start = vol
    vol_min = vfrac * vol_start

    # initialite filter
    if do_filter:
        if neighbours is None:
            neighbours = cells_around(centers, r_max, frmt='dict')
        if isinstance(neighbours, csr):
            factors = get_filter_factors_csr(centers, neighbours, r_max)
            fltr = sensitivity_filter_csr
        else:
            factors = get_filter_factors(centers, neighbours, r_max)
            fltr = sensitivity_filter

    # initialize penalty parameters
    if penalty is not None:
        p_start = penalty
        p_stop = penalty + 1
        p_step = maxiter + 1

    # ------------- INITIAL FEASIBLE SOLUTION ------------
    if guess is None:
        dens[:] = vfrac
    else:
        dens = deepcopy(guess)
    vol = np.sum(dens * vols)
    comp = compliance(update_stiffness=True)
    cIter = i_start
    p = p_start
    yield OptRes(dens, comp, vol, p, cIter)

    # ------------------- ITERATION -------------------
    dt = 0
    terminate = False
    while not terminate:
        if (p < p_stop) and (np.mod(cIter, p_step) == 0):
            p += p_inc
        
        # estimate lagrangian
        lagr = p * comp / vol

        # set up boundaries of change
        _dens = dens * (1 - dtol)
        np.clip(_dens, 1e-5, 1.0, out=_dens)
        dens_ = dens * (1 + dtol)
        np.clip(dens_, 1e-5, 1.0, out=dens_)

        # sensitivity [*(-1)]
        dCdx[:] = p * comps * dens ** (p-1)

        # sensitivity filter
        if do_filter:
            dCdx[:] = fltr(dCdx, dens, neighbours, factors)

        # calculate new densities and lagrangian
        dens_tmp_[:] = dens * (dCdx / vols) ** q
        dens_tmp[:] = dens_tmp_
        _lagr = 0
        lagr_ = 2 * lagr
        _maxtries = 200
        _ntries = 0
        while (lagr_ - _lagr) > 1e-3:
            if _ntries == _maxtries:
                raise RuntimeError("Couldn't find multiplier :(.")
            _lagr_ = (_lagr + lagr_) / 2
            dens_tmp[:] = dens_tmp_ / (_lagr_ ** q)
            np.clip(dens_tmp, _dens, dens_, out=dens_tmp)
            vol_tmp = np.sum(dens_tmp * vols)
            if vol_tmp < vol_min:
                lagr_ = _lagr_
            else:
                _lagr = _lagr_
            _ntries += 1
        lagr = lagr_
        dens[:] = dens_tmp

        # resolve equilibrium equations and calculate compliance
        comp = compliance(update_stiffness=True)
        dt += femsolver._summary['proc', 'time']
        vol = np.sum(dens * vols)
        cIter += 1
        res = OptRes(dens, comp, vol, p, cIter)
        yield res

        if nostop:
            terminate = False
        else:
            if cIter < miniter:
                terminate = False
            elif cIter >= maxiter:
                terminate = True
            else:
                terminate = (p >= p_stop)
        
    femsolver.K[:, :, :] = K_bulk_0
    return OptRes(dens, comp, vol, p, cIter)



# -*- coding: utf-8 -*-

try:
    import pypardiso
    __haspardiso__ = True
except Exception:
    __haspardiso__ = False
    

try:
    import sympy
    __hassympy__ = True
except Exception:
    __hassympy__ = False
    
    
try:
    import pylatex
    __haspylatex__ = True
except Exception:
    __haspylatex__ = False



# -*- coding: utf-8 -*-
from enum import Enum
from typing import Iterable, List


class DOF(Enum):
    """
    Degrees of freedom.
    
    Examples
    --------
    >>> from sigmaepsilon.solid.fem.config import DOF
    >>> DOF['UXX']
    <DOF.ROTX: 3>
    >>> DOF.UXX
    <DOF.ROTX: 3>
    >>> DOF.UXX.value
    3
    >>> DOF.UXX.name
    'ROTX'
    
    """
    UX = 0
    UY = 1
    UZ = 2
    ROTX = UXX = UYZ = 3
    ROTY = UYY = UXZ = 4
    ROTZ = UZZ = UXY = 5
    
    @classmethod
    def dofmap(cls, dofs:Iterable) -> List:
        return [cls[d].value for d in dofs]
    

if __name__=='__main__':

    DOF.UXX, DOF['UXX']
    type(DOF['UXX'].value)



# -*- coding: utf-8 -*-
import numpy as np

from sectionproperties.pre.pre import DEFAULT_MATERIAL, Material
from sectionproperties.pre.geometry import Geometry
from sectionproperties.pre.library.steel_sections import circular_hollow_section as CHS
from sectionproperties.pre.library.steel_sections import rectangular_hollow_section as RHS
from sectionproperties.pre.library.primitive_sections import rectangular_section as RS
from sectionproperties.pre.library.steel_sections import i_section
from sectionproperties.pre.library.steel_sections import tapered_flange_i_section as TFI
from sectionproperties.pre.library.steel_sections import channel_section as PFC
from sectionproperties.pre.library.steel_sections import tapered_flange_channel as TFC
from sectionproperties.pre.library.steel_sections import tee_section
from sectionproperties.analysis.section import Section

from dewloosh.core.wrapping import Wrapper
from linkeddeepdict.tools.kwargtools import getallfromkwargs

from neumann.array import atleastnd, ascont
from polymesh.utils import centralize
from polymesh.tri.trimesh import TriMesh
from polymesh.tet.tetmesh import TetMesh
from polymesh.topo.tr import T6_to_T3
from polymesh.topo import detach_mesh_bulk

from .utils import calc_beam_stresses_2d, calc_beam_stresses_4d


def generate_mesh(geometry: Geometry, *, l_max: float = None, a_max: float = None, 
                  n_max: int = None) -> Geometry:
    """
    Calculates a float describing the maximum mesh element area to be used
    for the finite-element mesh of the section. 

    Parameters
    ----------
    geometry : :class:`sectionproperties.pre.geometry.Geometry`
        Describes the shape of the section, i.e. 'I', 'H', 'RHS', etc.

    l_max : float, Optional
        Maximum element edge length. Default is None.

    a_max : float, Optional
        Maximum element area. Default is None.

    n_max : int, Optional
        Maximum number of elements. Default is None.

    Note
    ----
    The value of the mesh size is derived from the assumption of a perfectly 
    regular mesh of equilateral triangles.

    Returns
    ------- 
    :class:`sectionproperties.pre.geometry.Geometry`
        The geometry object provided with a finite element mesh.

    """
    area = geometry.calculate_area()
    mesh_sizes_max = []
    if isinstance(l_max, float):
        mesh_sizes_max.append(l_max**2 * np.sqrt(3) / 4)
    if isinstance(a_max, float):
        mesh_sizes_max.append(a_max)
    if isinstance(n_max, int):
        mesh_sizes_max.append(area / n_max)

    mesh_size_max = None
    if len(mesh_sizes_max) > 0:
        mesh_size_max = min(mesh_sizes_max)

    geometry.create_mesh(mesh_sizes=[mesh_size_max])
    return geometry


def get_section(shape, *args, mesh_params: dict = None, material: Material = None,
                **section_params) -> Section:
    """
    Returns a :class:`sectionproperties.analysis.section.Section` instance.

    The parameters in `section_params` are forwarded to the appropriate constructor of the 
    `sectionproperties` library. For the available parameters, see their documentation:

        https://sectionproperties.readthedocs.io/en/latest/rst/section_library.html

    Parameters
    ----------
    shape : str
        Describes the shape of the section. The currently supported section types
        are 'I', 'CHS', 'RHS', 'TFI', 'PFC', 'TFC', 'T'.

    mesh_params : dict, Optional
        A dictionary of parameters controlling mesh generation. Default is None.
        For the possible keys and values see :func:`generate_mesh`, to which
        these parameters are forwarded to. Default is None.
        
    material : :class:`sectionproperties.pre.pre.Material`, Optional
        The material of the section. If not specified, a default material is
        used. Default is None. 

    **section_params : dict
        Parameters required for a given section. See the documentation of
        the `sectionproperties` library for more details. The parameters
        are only required, if the first argument is a string.

    Note
    ----
    Specification of a material is only necessary, if stresses are to be
    calculated. If the reason of creating the section is the calculation of
    geometrical properties of a section, the choice of material is irrelevant.
    
    Returns
    -------
    :class:`sectionproperties.analysis.section.Section`
        An object representing a cross-section of a beam.

    Examples
    --------
    >>> from sigmaepsilon.solid.model.bernoulli import get_section
    >>> mesh_params = dict(n_min=100, n_max=500)
    >>> section = get_section('CHS', d=1.0, t=0.1, n=64, mesh_params=mesh_params)

    """

    material = DEFAULT_MATERIAL if material is None else material
    geom, ms = None, None
    if shape == 'CHS':
        geom = CHS(d=section_params['d'], t=section_params['t'],
                   n=section_params.get('n', 64), material=material)
        ms = section_params['t']
    elif shape == 'RS':
        keys = ['d', 'b']
        params = getallfromkwargs(keys, **section_params)
        geom = RS(material=material, **params)
        ms = section_params['t']
    elif shape == 'RHS':
        keys = ['d', 'b', 't', 'n_out', 'n_r']
        params = getallfromkwargs(keys, **section_params)
        geom = RHS(material=material, **params)
        ms = section_params['t']
    elif shape == 'I':
        keys = ['d', 'b', 't_f', 't_w', 'r', 'n_r']
        params = getallfromkwargs(keys, **section_params)
        geom = i_section(material=material, **params)
        ms = min(section_params['t_f'], section_params['t_w'])
    elif shape == 'TFI':
        keys = ['d', 'b', 't_f', 't_w', 'r_r', 'r_f', 'alpha', 'n_r']
        params = getallfromkwargs(keys, **section_params)
        geom = TFI(material=material, **params)
        ms = min(section_params['t_f'], section_params['t_w'])
    elif shape == 'PFC':
        keys = ['d', 'b', 't_f', 't_w', 'r', 'n_r']
        params = getallfromkwargs(keys, **section_params)
        geom = PFC(material=material, **params)
        ms = min(section_params['t_f'], section_params['t_w'])
    elif shape == 'TFC':
        keys = ['d', 'b', 't_f', 't_w', 'r_r', 'r_f', 'alpha', 'n_r']
        params = getallfromkwargs(keys, **section_params)
        geom = TFC(material=material, **params)
        ms = min(section_params['t_f'], section_params['t_w'])
    elif shape == 'T':
        keys = ['d', 'b', 't_f', 't_w', 'r', 'n_r']
        params = getallfromkwargs(keys, **section_params)
        geom = tee_section(material=material, **params)
        ms = min(section_params['t_f'], section_params['t_w'])
    else:
        raise NotImplementedError(
            "Section type <{}> is not yet implemented :(".format(shape))

    if geom is not None:
        if mesh_params is None:
            assert ms is not None, "Invalid input!"
            mesh_params = dict(l_max=ms)
        assert isinstance(mesh_params, dict)
        geom = generate_mesh(geom, **mesh_params)
        return Section(geom)
    raise RuntimeError("Unable to get section.")


class BeamSection(Wrapper):
    """
    Wraps an instance of `sectionproperties.analysis.section.Section` and
    adds a little here and there to make some of the functionality more
    accessible.

    Parameters
    ----------
    *args : tuple, Optional
        The first parameter can be a string referring to a section type, or an
        instance of :class:`sectionproperties.analysis.section.Section`. In the
        former case, parameters of the section can be provided as keyword arguments,
        which are then forwarded to :func:`get_section`, see its documentation
        for further details.
        
    **kwargs : dict, Optional
        The parameters of a section as keyword arguments, only if the first
        positional argument is a string (see above). 
         
    mesh_params : dict, Optional
        A dictionary controlling the density of the mesh of the section.
        Default is None.
        
    material : :class:`sectionproperties.pre.pre.Material`, Optional
        The material of the section. If not specified, a default material is
        used. Default is None.
        
    wrap : :class:`sectionproperties.analysis.section.Section`
        A section object to be wrapped. It can also be provided as the
        first positional argument. 

    Notes
    -----
    The implementation here only covers homogeneous cross sections. If you want
    to define an inhomogeneous section, it must be explicity given either as
    the first position argument, or with the keyword argument `wrap`.

    Examples
    --------
    >>> from sigmaepsilon.solid import BeamSection
    >>> section = BeamSection(get_section('CHS', d=1.0, t=0.1, n=64))

    or simply provide the shape as the first argument and everything 
    else with keyword arguments:

    >>> section = BeamSection('CHS', d=1.0, t=0.1, n=64)

    """

    def __init__(self, *args, wrap=None, shape=None, mesh_params=None, 
                 material: Material = None, **kwargs):
        if len(args) > 0:
            try:
                if isinstance(args[0], str):
                    wrap = get_section(
                        args[0], mesh_params=mesh_params, material=material, **kwargs)
                else:
                    if shape is None:
                        if isinstance(args[0], Section):
                            wrap = args[0]
                    else:
                        wrap = get_section(
                            shape, mesh_params=mesh_params, material=material, **kwargs)
            except Exception:
                raise RuntimeError("Invalid input.")
        super().__init__(*args, wrap=wrap, **kwargs)
        self.props = None

    def coords(self) -> np.ndarray:
        """
        Returns centralized vertex coordinates of the supporting 
        point cloud as a numpy array.
        """
        return centralize(np.array(self.mesh['vertices']))

    def topology(self) -> np.ndarray:
        """
        Returns vertex indices of T6 triangles as a numpy array.
        """
        return np.array(self.mesh['triangles'].tolist())

    def trimesh(self, subdivide=False, T6=False, **kwargs) -> TriMesh:
        """
        Returns the mesh of the section as a collection of T3 triangles.
        Keyword arguments are forwarded to the constructor of 
        :class:`polymesh.tri.trimesh.TriMesh`.

        Parameters
        ----------
        T6 : boolean, Optional
            If False, the original T6 trinagles are transformed into T3
            triangles. Default is False.

        subdivide : boolean, Optional
            Controls how the T6 triangles are transformed into T3 triangles,
            if the argument 'T6' is False. If True, the T6 triangles
            are subdivided into 4 T3 triangles. If False, T3 triangles are
            formed by the corners of T6 triangles only, and all the remaining
            nodes are neglected.

        See Also
        --------
        :class:`polymesh.tri.trimesh.TriMesh`

        Examples
        --------
        >>> from sigmaepsilon.solid import BeamSection
        >>> section = BeamSection(get_section('CHS', d=1.0, t=0.1, n=64))
        >>> trimesh = section.trimesh()

        """
        points, triangles = self.coords(), self.topology()
        if not T6:
            if subdivide:
                path = np.array([[0, 5, 4], [5, 1, 3],
                                [3, 2, 4], [5, 3, 4]], dtype=int)
                points, triangles = T6_to_T3(points, triangles, path=path)
            else:
                points, triangles = detach_mesh_bulk(points, triangles[:, :3])
        return TriMesh(points=points, triangles=triangles, **kwargs)

    def extrude(self, *args, length=None, frame=None, N=None, **kwargs) -> TetMesh:
        """
        Creates a 3d tetragonal mesh from the section.

        Parameters
        ----------
        length : float
            Length of the beam.

        N : int
            Number of subdivisions along the length of the beam.

        frame : numpy.ndarray
            A 3x3 matrix representing an orthonormal coordinate frame.

        """
        return self.trimesh(frame=frame).extrude(h=length, N=N)

    def calculate_geometric_properties(self, *args, **kwargs):
        return self._wrapped.calculate_geometric_properties(*args, **kwargs)

    def calculate_warping_properties(self, *args, **kwargs):
        return self._wrapped.calculate_warping_properties(*args, **kwargs)

    @property
    def A(self):
        """
        Returns the cross-sectional area.
        """
        return self.section_props.area

    @property
    def Ix(self):
        """
        Returns the second moment of inertia around 'x'.
        """
        return self.Iy + self.Iz

    @property
    def Iy(self):
        """
        Returns the second moment of inertia around 'y'.
        """
        return self.section_props.ixx_c

    @property
    def Iz(self):
        """
        Returns the second moment of inertia around 'z'.
        """
        return self.section_props.iyy_c

    @property
    def geometric_properties(self):
        """
        Returns the geometric properties of the section.
        """
        return {'A': self.A, 'Ix': self.Ix, 'Iy': self.Iy, 'Iz': self.Iz}

    @property
    def section_properties(self):
        """
        Returns all properties of the section.
        """
        return self.props

    def calculate_section_properties(self, separate=False):
        """
        Calculates all properties of the section.
        """
        if not separate:
            # 0-xx 1-yy 2-zz 3-yz 4-xz 5-xy
            strs = np.zeros((6, 6, len(self.mesh['vertices'])))
            """strs = {'xx': {}, 'yy': {}, 'zz': {}, 'xy': {}, 'xz': {}, 'yz': {}}"""
        else:
            raise NotImplementedError

        self.calculate_geometric_properties()
        section_properties = self.geometric_properties

        self.calculate_warping_properties()
        _strs = np.eye(6)
        for i in range(6):
            N, Vy, Vx, Mzz, Mxx, Myy = _strs[i]
            stress_post = self.calculate_stress(
                N=N, Vy=Vy, Vx=Vx, Mzz=Mzz, Mxx=Mxx, Myy=Myy
            )
            stresses = stress_post.get_stress()
            if not separate:
                # this assumes that the section is homogeneous, hence the 0 index
                strs[0, i, :] = stresses[0]['sig_zz']  # xx
                strs[5, i, :] = stresses[0]['sig_zx']  # xy
                strs[4, i, :] = stresses[0]['sig_zy']  # xz
            else:
                raise NotImplementedError

        section_properties['stress-factors'] = strs
        self.props = section_properties
        return section_properties

    def get_section_properties(self):
        """
        Returns all properties of the section.
        """
        if self.props is not None:
            return self.props

    def model_stiffness_matrix(self, E=None, nu=None) -> np.ndarray:
        """
        Returns the model stiffness matrix of the section.

        Parameters
        ----------
        E : float, Optional
            Young's modulus. If not provided, the value is obtained
            from the associated material of the section, if there is one.
            Default is None.

        nu : float, Optional
            Poisson's ratio. If not provided, the value is obtained
            from the associated material of the section, if there is one.
            Default is None.

        Returns
        -------
        numpy.ndarray
            The 4x4 elastic ABDS matrix of the section.

        """
        if E is None:
            E = self.geometry.material.elastic_modulus
            nu = self.geometry.material.poissons_ratio
        G = E / (2 * (1 + nu))
        return np.array([
            [E*self.A, 0, 0, 0],
            [0, G*self.Ix, 0, 0],
            [0, 0, E*self.Iy, 0],
            [0, 0, 0, E*self.Iz]
        ])

    def calculate_stresses(self, forces: np.ndarray) -> np.ndarray:
        """
        Returns stresses for all points of the mesh. The internal forces
        are expected in the order Fx, Fy, Fz, Mx, My, Mz, where x, y and z 
        denote the local axes of the section, the stress components are
        returned in the order s11, s22, s33, s12, s13, s23.

        Parameters
        ----------
        forces : numpy.ndarray    
            It can be either

                A 1d float array representing all internal forces of a beam.

                A 2d float array. In this case it is assumed, that the provided
                data has the shape (nX, nSTRE), where 'nX' is the number of records
                in the data.

                A 3d float array. In this case it is assumed, that the provided
                data has the shape (nE, nP, nSTRE), representing
                all internal forces (nSTRE) for multiple elements(nE) and 
                evaluation points (nP).

                A 4d float array of shape (nE, nP, nSTRE, nRHS), representing
                all internal forces (nSTRE) for multiple elements(nE), load cases(nRHS) 
                and evaluation points (nP).

        Returns
        -------
        numpy.ndarray
            A numpy float array that can be

                A 2d array of shape (nPS, nS),

                A 5d array of shape (nE, nP, nPS, nS, nRHS),

            where

                nPS : the number of points of the mesh of the section

                nS : the number of stress components

                nE : the number of finite elements

                nP : the number of evaluation points for an element

                nRHS : the number of right hand sides (possibly load cases)

        Notes
        -----
        This only covers homogeneous cross sections.

        """
        assert isinstance(forces, np.ndarray)
        factors = self.props['stress-factors']
        nD = len(forces.shape)
        if nD == 1:
            f_ = ascont(atleastnd(forces, n=2, front=True))
            return calc_beam_stresses_2d(f_, factors)[0, ...]
        elif nD == 2:
            return calc_beam_stresses_2d(forces, factors)
        elif nD == 3:
            f_ = ascont(atleastnd(forces, n=4, back=True))
            return calc_beam_stresses_4d(f_, factors)[..., 0]
        elif nD == 4:
            return calc_beam_stresses_4d(forces, factors)
        else:
            raise RuntimeError("Unrecognizable shape of internal forces.")



# -*- coding: utf-8 -*-
from typing import Union

from numba import njit, prange
import numpy as np
from numpy import ndarray
from scipy.sparse import spmatrix

from neumann.linalg.sparse.csr import csr_matrix as csr
from neumann.array import find1d, flatten2dC

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def fixity2d_to_dofs1d(fixity2d: np.ndarray, inds: np.ndarray = None):
    """
    Returns the indices of the degrees of freedoms
    being supressed. 

    Optionally, global indices of the rows in 'fixity2d' 
    array can be provided by the optional argument 'inds'.

    Parameters
    ----------
    fixity2d : numpy.ndarray
        2d numpy array of booleans. It has as many rows as nodes, and as 
        many columns as derees of freedom per node in the model. A True
        value means that the corresponding dof is fully supressed.

    inds : numpy.ndarray, Optional
        1d numpy array of integers, default is None. If provided, it should
        list the global indices of the nodes, for which the data is provided
        by the array 'fixity2d'.

    Returns
    -------
    dofinds : np.ndarray
        1d numpy array of integers. 

    """
    ndof = fixity2d.shape[1]
    args = np.argwhere(fixity2d == True)
    N = args.shape[0]
    res = np.zeros(N, dtype=args.dtype)
    for i in prange(N):
        res[i] = args[i, 0]*ndof + args[i, 1]
    if inds is None:
        return res
    else:
        dofinds = np.zeros_like(res)
        for i in prange(len(inds)):
            for j in prange(ndof):
                dofinds[i*ndof + j] = inds[i]*ndof + j
        return dofinds[res]


@njit(nogil=True, parallel=True, cache=__cache)
def nodes2d_to_dofs1d(inds: np.ndarray, values: np.ndarray):
    """
    Returns a tuple of degree of freedom indices and data, 
    based on a nodal definition.

    Parameters
    ----------
    inds : numpy.ndarray
        1d numpy array of integers, listing global node indices.

    values : numpy.ndarray
        3d numpy float array of shape (nN, nDOF, nX), listing values 
        for each node in 'inds'. nX is the number of datasets.

    Returns
    -------
    dofinds : numpy.ndarray
        1d numpy array of integers, denoting global dof indices.

    dofvals : numpy.ndarray
        2d numpy float array of shape (nN * nDOF, nX), denoting values 
        on dofs in 'dofinds'.

    """
    nN, dof_per_node, nRHS = values.shape
    dofinds = np.zeros((nN * dof_per_node), dtype=inds.dtype)
    dofvals = np.zeros((nN * dof_per_node, nRHS), dtype=values.dtype)
    for node in prange(nN):
        for dof in prange(dof_per_node):
            i = node * dof_per_node + dof
            dofinds[i] = inds[node] * dof_per_node + dof
            for rhs in prange(nRHS):
                dofvals[i, rhs] = values[node, dof, rhs]
    return dofinds, dofvals


@njit(nogil=True, cache=__cache, parallel=True)
def weighted_stiffness_bulk(K: np.ndarray, weights: np.ndarray):
    """
    Returns a weighted stiffness matrix.

    Parameters
    ----------
    K : numpy.ndarray
        2d numpy array of floats, where the first axis of K runs 
        along the elements.

    weights : numpy.ndarray
        1d numpy array of floats.

    Returns
    -------
    Kw : numpy.ndarray
        2d numpy array of floats with the same shape as 'K'.

    """
    nE = len(weights)
    res = np.zeros_like(K)
    for i in prange(nE):
        res[i, :] = weights[i] * K[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def irows_icols_bulk(edofs: np.ndarray):
    """
    Returns row and column index data for several finite elements.

    Parameters
    ----------
    edofs : numpy.ndarray
        2d numpy array. Each row has the meaning of global degree of 
        freedom numbering for a given finite element.

    Returns
    -------
    irows : numpy.ndarray
        Global indices of the rows of the entries of stiffness matrices
        of the elements.
        
    icols : numpy.ndarray
        Global indices of the columns of the entries of stiffness matrices
        of the elements.

    Notes
    -----
    The implementation assumes that every cell has the same number of
    degrees of freedom.

    """
    nE, nNE = edofs.shape
    nTOTV = nNE * nNE
    irows = np.zeros((nE, nTOTV), dtype=edofs.dtype)
    icols = np.zeros((nE, nTOTV), dtype=edofs.dtype)
    for iNE in prange(nNE):
        for jNE in prange(nNE):
            i = iNE * nNE + jNE
            for iE in prange(nE):
                irows[iE, i] = edofs[iE, iNE]
                icols[iE, i] = edofs[iE, jNE]
    return irows, icols


@njit(nogil=True, cache=__cache, parallel=True)
def irows_icols_bulk_filtered(edofs: np.ndarray, inds: np.ndarray):
    """
    Returns row and column index data for finite elements specified
    by the index array `inds`.

    Parameters
    ----------
    edofs : numpy.ndarray
        2d numpy array. Each row has the meaning of global degree of 
        freedom numbering for a given finite element.

    inds: numpy.ndarray
        1d numpy array of integers specifying active elements in an assembly.

    Returns
    -------
    irows : numpy.ndarray
        Global indices of the rows of the entries of stiffness matrices
        of the elements.
        
    icols : numpy.ndarray
        Global indices of the columns of the entries of stiffness matrices
        of the elements.

    Notes
    -----
    The implementation assumes that every cell has the same number of
    degrees of freedom.

    """
    nI = len(inds)
    nNE = edofs.shape[1]
    nTOTV = nNE * nNE
    irows = np.zeros((nI, nTOTV), dtype=edofs.dtype)
    icols = np.zeros((nI, nTOTV), dtype=edofs.dtype)
    for iNE in prange(nNE):
        for jNE in prange(nNE):
            i = iNE * nNE + jNE
            for iI in prange(nI):
                irows[iI, i] = edofs[inds[iI], iNE]
                icols[iI, i] = edofs[inds[iI], jNE]
    return irows, icols


@njit(nogil=True, cache=__cache, parallel=True)
def topo_to_gnum(topo: np.ndarray, ndofn: int):
    """
    Returns global dof numbering based on element 
    topology data.

    Parameters
    ----------
    topo : numpy.ndarray
        2d numpy array of integers. Topology array listing global
        node numbers for several elements.

    ndofn : int
        Number of degrees of freedoms per node.

    Returns
    -------
    gnum : numpy.ndarray
        2d numpy array of integers.

    """
    nE, nNE = topo.shape
    gnum = np.zeros((nE, nNE*ndofn), dtype=topo.dtype)
    for i in prange(nE):
        for j in prange(nNE):
            for k in prange(ndofn):
                gnum[i, j*ndofn + k] = topo[i, j]*ndofn + k
    return gnum


@njit(nogil=True, cache=__cache, parallel=True)
def topo_to_gnum_jagged(topo, gnum, ndofn: int):
    """
    Returns global dof numbering based on element 
    topology data.

    Parameters
    ----------
    topo : numpy.ndarray
        2d array of integers. Topology array listing global
        node numbers for several elements.

    gnum : numpy.ndarray
        2d buffer array for the results.

    ndofn : int
        Number of degrees of freedoms per node.

    Returns
    -------
    gnum : numpy.ndarray
        2d numpy array of integers.

    """
    nE, _ = topo.shape
    for i in prange(nE):
        for j in prange(len(topo[i])):
            for k in prange(ndofn):
                gnum[i, j*ndofn + k] = topo[i, j]*ndofn + k
    return gnum


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def penalty_factor_matrix(cellfixity: ndarray, shp: ndarray):
    nE, nNE, nDOF = cellfixity.shape
    N = nDOF * nNE
    res = np.zeros((nE, N, N), dtype=shp.dtype)
    for iE in prange(nE):
        for iNE in prange(nNE):
            for iDOF in prange(nDOF):
                fix = cellfixity[iE, iNE, iDOF]
                for jNE in prange(nNE):
                    i = jNE * nDOF + iDOF
                    for kNE in prange(nNE):
                        j = kNE * nDOF + iDOF
                        res[iE, i, j] += fix * shp[iNE, jNE] * shp[iNE, kNE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def approximation_matrix(ndf: ndarray, NDOFN: int):
    """
    Returns a matrix of approximation coefficients 
    for all elements.
    """
    nE, nNE = ndf.shape[:2]
    N = nNE * NDOFN
    nappr = np.eye(N, dtype=ndf.dtype)
    res = np.zeros((nE, N, N), dtype=ndf.dtype)
    for iE in prange(nE):
        for i in prange(nNE):
            for j in prange(nNE):
                for ii in prange(NDOFN):
                    for jj in prange(NDOFN):
                        res[iE, i*2 + ii, j*2 + jj] = nappr[i, j] * ndf[iE, i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def nodal_approximation_matrix(ndf: ndarray):
    """
    Returns a matrix of nodal approximation coefficients 
    for all elements.
    """
    nE, nNE = ndf.shape[:2]
    nappr = np.eye(nNE, dtype=ndf.dtype)
    res = np.zeros((nE, nNE, nNE), dtype=ndf.dtype)
    for iE in prange(nE):
        for i in prange(nNE):
            for j in prange(nNE):
                res[iE, i, j] = nappr[i, j] * ndf[iE, i]
    return res


@njit(nogil=True, cache=__cache)
def nodal_compatibility_factors(nam_csr_tot: csr, nam: ndarray, topo: ndarray):
    indptr = nam_csr_tot.indptr
    indices = nam_csr_tot.indices
    data = nam_csr_tot.data
    nN = len(indptr)-1
    widths = np.zeros(nN, dtype=indptr.dtype)
    for iN in range(nN):
        widths[iN] = indptr[iN+1] - indptr[iN]
    factors = dict()
    nreg = dict()
    for iN in range(nN):
        factors[iN] = np.zeros((widths[iN], widths[iN]), dtype=nam.dtype)
        nreg[iN] = indices[indptr[iN]: indptr[iN+1]]
    nE, nNE = topo.shape
    for iE in range(nE):
        topoE = topo[iE]
        for jNE in range(nNE):
            nID = topo[iE, jNE]
            dataN = data[indptr[nID]: indptr[nID+1]]
            topoN = nreg[nID]
            dataNE = np.zeros(widths[nID], dtype=nam.dtype)
            imap = find1d(topoE, topoN)
            dataNE[imap] = nam[iE, jNE]
            rNE = dataN - dataNE  # residual
            factors[nID] += np.outer(rNE, rNE)
    return factors, nreg


@njit(nogil=True, parallel=True, cache=__cache)
def compatibility_factors_to_coo(ncf: dict, nreg: dict):
    """
    ncf : nodal_compatibility_factors
    """
    nN = len(ncf)
    widths = np.zeros(nN, dtype=np.int32)
    for iN in prange(nN):
        widths[iN] = len(nreg[iN])
    shapes = (widths**2).astype(np.int64)
    N = np.sum(shapes)
    data = np.zeros(N, dtype=np.float64)
    rows = np.zeros(N, dtype=np.int32)
    cols = np.zeros(N, dtype=np.int32)
    c = 0
    for iN in range(nN):
        data[c: c + shapes[iN]] = flatten2dC(ncf[iN])
        nNN = widths[iN]
        for jNN in prange(nNN):
            for kNN in prange(nNN):
                rows[c + jNN*nNN + kNN] = nreg[iN][jNN]
                cols[c + jNN*nNN + kNN] = nreg[iN][kNN]
        c += shapes[iN]
    return data, rows, cols


@njit(nogil=True, parallel=True, cache=__cache)
def topo1d_to_gnum1d(topo1d: ndarray, NDOFN: int):
    nN = topo1d.shape[0]
    gnum1d = np.zeros(nN * NDOFN, dtype=topo1d.dtype)
    for i in prange(nN):
        for j in prange(NDOFN):
            gnum1d[i*NDOFN + j] = topo1d[i] * NDOFN + j
    return gnum1d


@njit(nogil=True, parallel=True, cache=__cache)
def ncf_to_cf(ncf: ndarray, NDOFN: int):
    nN = ncf.shape[0]
    cf = np.zeros((nN * NDOFN, nN * NDOFN), dtype=ncf.dtype)
    for i in prange(nN):
        for ii in prange(NDOFN):
            for j in prange(nN):
                cf[i*NDOFN + ii, j*NDOFN + ii] = ncf[i, j]
    return cf


@njit(nogil=True, cache=__cache)
def compatibility_factors(ncf: dict, nreg: dict, NDOFN: int):
    """ncf : nodal_compatibility_factors"""
    nN = len(ncf)
    widths = np.zeros(nN, dtype=np.int32)
    for iN in prange(nN):
        widths[iN] = len(nreg[iN])
    cf = dict()
    reg = dict()
    for iN in range(nN):
        cf[iN] = ncf_to_cf(ncf[iN], NDOFN)
        reg[iN] = topo1d_to_gnum1d(nreg[iN], NDOFN)
    return cf, reg


@njit(nogil=True, parallel=True, cache=__cache)
def element_dofmap_bulk(dofmap: ndarray, nDOF: int, nNODE: int):
    nDOF_ = dofmap.shape[0]
    res = np.zeros(nNODE * nDOF_, dtype=dofmap.dtype)
    for i in prange(nNODE):
        for j in prange(nDOF_):
            res[i*nDOF_ + j] = i*nDOF + dofmap[j]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def expand_coeff_matrix_bulk(A_in: ndarray, A_out: ndarray, dofmap: ndarray):
    """
    Expands the local coefficient matrices into a standard form. It is used
    for instance when the total matrix is built up of smaller parts.

    """
    nDOF = dofmap.shape[0]
    for i in prange(nDOF):
        for j in prange(nDOF):
            A_out[:, dofmap[i], dofmap[j]] = A_in[:, i, j]
    return A_out


@njit(nogil=True, parallel=True, cache=__cache)
def nodal_mass_matrix_data(nodal_masses: ndarray, ndof: int = 6):
    N = nodal_masses.shape[0]
    res = np.zeros((N*ndof))
    for i in prange(N):
        res[i*ndof: i*ndof + 3] = nodal_masses[i]
    return res


def min_stiffness_diagonal(K: Union[ndarray, spmatrix]) -> float:
    """
    Returns the minimum diagonal entry of the stiffness matrix.
    
    Parameters
    ----------
    K : Union[ndarray, spmatrix]
        The stiffness matrix as a dense or a sparse array.
        
    """
    if isinstance(K, spmatrix):
        return np.min(K.diagonal())
    else:
        i = np.arange(K.shape[-1])
        return np.min(K[:, i, i])



# -*- coding: utf-8 -*-
from typing import Tuple, Union, List

import numpy as np
from numpy import ndarray
from scipy.sparse import coo_matrix

from dewloosh.core.wrapping import Wrapper
from neumann import squeeze
from neumann.array import repeat

from .mesh import FemMesh
from .femsolver import StaticSolver, DynamicSolver
from .preproc import assemble_load_vector
from .tr import tr_element_matrices_bulk, tr_nodal_loads_bulk
from .dyn import effective_modal_masses, Rayleigh_quotient
from .pointdata import flatten_pd
from .constants import DEFAULT_DIRICHLET_PENALTY, DEFAULT_MASS_PENALTY_RATIO

__all__ = ['Structure']


class Structure(Wrapper):
    """
    A higher level class to manage solid structures.

    Parameters
    ----------
    mesh : FemMesh, Optional
        A finite element mesh.   
    essential_penalty : float, Optional
        Penalty parameter for the Courant penalty function used to enforce
        Dirichlet (essential) boundary conditions. If not provided, a default
        value of `sigmaepsilon.solid.fem.constants.DEFAULT_DIRICHLET_PENALTY` is used.
    mass_penalty_ratio : float, Optional
        Ratio of the penalty factors applied to the stiffness matrix (pK) and
        the mass matrix (pM) as pM/pK. If not provided, a default
        value of `sigmaepsilon.solid.fem.constants.DEFAULT_MASS_PENALTY_RATIO` is used.

    """

    def __init__(self, 
                 mesh: FemMesh = None, 
                 essential_penalty: float=DEFAULT_DIRICHLET_PENALTY,
                 mass_penalty_ratio: float=DEFAULT_MASS_PENALTY_RATIO
                 ):
        if not isinstance(mesh, FemMesh):
            raise TypeError("The type of 'mesh' is {},"
                            " which is not a subclass "
                            "of 'FemMesh'".format(type(mesh)))
        if not mesh.locked:
            mesh.lock(create_mappers=True)
        super().__init__(wrap=mesh)
        self._static_solver = None
        self._dynamic_solver_ = None
        self._natural_circular_frequencies = None
        self._essential_penalty = essential_penalty
        self._mass_penalty_ratio = mass_penalty_ratio
        self.clear_constraints()

    @property
    def mesh(self) -> FemMesh:
        """
        Returns the underlying mesh object.
        """
        return self._wrapped

    @mesh.setter
    def mesh(self, value: FemMesh):
        """
        Sets the underlying mesh object.
        """
        if not isinstance(value, FemMesh):
            raise TypeError("The type of 'value' is {},"
                            " which is not a subclass "
                            "of 'FemMesh'".format(type(value)))
        if not value.locked:
            value.lock(create_mappers=True)
        self._wrapped = value

    @property
    def constraints(self) -> List:
        """
        Returns the constraints of the structure as a list. Note that
        this is an independent mechanism from the 'fixity' data defined
        for the :class:`sigmaepsilon.solid.fem.pointdata.PointData` object 
        of the underlying mesh instance.
        """
        if self._constraints is None:
            self._constraints = []
        return self._constraints

    def clear_constraints(self):
        """
        Clears up the constraints. This does not affect the 'fixity'
        definitions in the pointdata of the mesh.
        """
        self._constraints = None

    @flatten_pd(False)
    @squeeze(True)
    def natural_modes_of_vibration(self, *, flatten: bool = False,
                                   squeeze: bool = True, **kwargs) -> ndarray:
        """
        Returns natural modes of vibration.

        Parameters
        ----------
        flatten : bool, Optional
            If True, the result is a 1d array, otherwise 2d. Default is False.

        squeeze : bool, Optional
            Calls :func:`numpy.squeeze` on the result. This might be relevant for 
            single element operations, or if there is only one load case, etc. Then,
            it depends on the enviroment if a standard shape is desirable to maintain 
            or not. Default is True.

        Returns
        -------
        numpy.ndarray

        """
        return self.mesh.pd.vshapes

    def natural_circular_frequencies(self, *args, return_vectors: bool = False,
                                     **kwargs) -> Union[Tuple[ndarray], ndarray]:
        """
        Returns natural circular frequencies. The call forwards all parameters
        to :func:`sigmaepslion.solid.fem.dyn.natural_circular_frequencies`, 
        see the docs there for the details. 

        Parameters
        ----------
        return_vectors : bool, Optional
            To return eigenvectors or not. Default is False.

        Returns
        -------
        numpy.ndarray
            The natural circular frequencies.
        numpy.ndarray
            The eigenvectors, only if 'return_vectors' is True.

        See also
        --------
        :func:`sigmaepslion.solid.fem.dyn.natural_circular_frequencies`

        """
        f = self._dynamic_solver_.natural_circular_frequencies(*args, **kwargs)
        if return_vectors:
            v = self.natural_modes_of_vibration(flatten=True)
            return f, v
        return f

    def elastic_stiffness_matrix(self, *args, penalize: bool = False,
                                 sparse: bool = False, transform: bool = False,
                                 **kwargs) -> Union[ndarray, coo_matrix]:
        """
        Returns the elastic stiffness matrix of the structure with dense or 
        sparse layout.

        Parameters
        ----------
        penalize : bool, Optional
            If True, the result is the sum of the stiffness matrix and the
            penalty matrix of the essential boundary conditions.
        sparse : bool, Optional
            If True, the result is a sparse matrix. Default is False.
        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is True.

        Returns
        -------
        numpy.ndarray or scipy.sparse.coo_matrix

        """
        if penalize:
            assert transform, "Must transform to penalize."
            assert sparse, "Penalization is only available for sparse results."
        if sparse:
            assert transform, "Must transform for a sparse output."
        params = dict(sparse=sparse, transform=transform)
        params.update(**kwargs)
        M_sparse = self.mesh.elastic_stiffness_matrix(**params)
        if penalize:
            M_sparse += self.mesh.essential_penalty_matrix()
        return M_sparse.tocoo()

    def essential_penalty_matrix(self, *args, **kwargs) -> coo_matrix:
        """
        Returns the penalty stiffness matrix of the structure.

        Returns
        -------
        scipy.sparse.coo_matrix

        """
        return self.mesh.essential_penalty_matrix(*args, **kwargs).tocoo()

    def consistent_mass_matrix(self, *, penalize: bool = False,
                               sparse: bool = False, transform: bool = False,
                               **kwargs) -> Union[coo_matrix, ndarray]:
        """
        Returns the stiffness-consistent mass matrix of the structure.

        Parameters
        ----------
        penalize : bool, Optional
            If True, the result is the sum of the mass matrix and the
            penalty matrix of the essential boundary conditions.
        sparse : bool, Optional
            If True, the result is a sparse matrix. Default is False.
        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is True.

        Returns
        -------
        numpy.ndarray or scipy.sparse.coo_matrix

        """
        if penalize:
            assert transform, "Must transform to penalize."
            assert sparse, "Penalization is only available for sparse results."
        if sparse:
            assert transform, "Must transform for a sparse output."
        params = dict(sparse=sparse, transform=transform)
        params.update(**kwargs)
        return self.mesh.consistent_mass_matrix(**params)

    def nodal_mass_matrix(self) -> coo_matrix:
        """
        Returns the nodal mass matrix of the structure.

        Returns
        -------
        scipy.sparse.coo_matrix

        """
        return self.mesh.nodal_mass_matrix().tocoo()

    def mass_matrix(self, penalize: bool = False) -> coo_matrix:
        """
        Returns the total mass matrix of the structure, which is the
        sum of the stiffness-consistent mass matrix derived from the
        densities of the cells and the nodal masses.

        Parameters
        ----------
        penalize : bool, Optional
            If True, the result is the sum of the mass matrix and the
            penalty matrix emenating from the enforcing of the essential 
            boundary conditions using a Courant penalty function.

        Returns
        -------
        scipy.sparse.coo_matrix

        """
        M_sparse = self.mesh.mass_matrix()
        if penalize:
            M_sparse += self.mesh.essential_penalty_matrix()
        return M_sparse.tocoo()

    @flatten_pd(False)
    @squeeze(True)
    def nodal_dof_solution(self, *, flatten: bool = False,
                           squeeze: bool = True, **kwargs) -> ndarray:
        """
        Returns the vector of nodal displacements and optionally stores the result
        with a specified key.

        Parameters
        ----------
        flatten : bool, Optional
            If True, the result is a 1d array, otherwise 2d. Default is False.

        squeeze : bool, Optional
            Calls :func:`numpy.squeeze` on the result. This might be relevant for 
            single element operations, or if there is only one load case, etc. Then,
            it depends on the enviroment if a standard shape is desirable to maintain 
            or not. Default is True.

        Returns
        -------
        numpy.ndarray

        """
        return self.mesh.pd.dofsol

    @flatten_pd(False)
    @squeeze(True)
    def reaction_forces(self, *args, flatten: bool = False, squeeze: bool = True,
                        store: str = None, **kwargs) -> ndarray:
        """
        Returns the vector of reaction forces.

        Parameters
        ----------
        flatten : bool, Optional
            If True, the result is a 1d array, otherwise 2d. Default is False.

        squeeze : bool, Optional
            Calls :func:`numpy.squeeze` on the result. This might be relevant for 
            single element operations, or if there is only one load case, etc. Then,
            it depends on the enviroment if a standard shape is desirable to maintain 
            or not. Default is True.

        Returns
        -------
        numpy.ndarray

        """
        return self.mesh.pd.reactions

    @flatten_pd(False)
    @squeeze(True)
    def nodal_forces(self, *args, flatten: bool = False,
                     squeeze: bool = True, **kwargs) -> ndarray:
        """
        Returns the vector of nodal forces from all sources. This is calculated
        during static analysis.

        Parameters
        ----------
        flatten : bool, Optional
            If True, the result is a 1d array, otherwise 2d. Default is False.

        squeeze : bool, Optional
            Calls :func:`numpy.squeeze` on the result. This might be relevant for 
            single element operations, or if there is only one load case, etc. Then,
            it depends on the enviroment if a standard shape is desirable to maintain 
            or not. Default is True.

        """
        return self.mesh.pd.forces

    @squeeze(True)
    def internal_forces(self, *args, flatten: bool = False, squeeze: bool = True,
                        **kwargs) -> ndarray:
        """
        Returns the internal forces for one or more elements.
        """
        return self.mesh.internal_forces(*args, flatten=flatten,
                                         squeeze=False, **kwargs)

    @squeeze(True)
    def external_forces(self, *args, flatten: bool = False, **kwargs) -> ndarray:
        """
        Returns the external forces for one or more elements.
        """
        return self.mesh.external_forces(*args, flatten=flatten,
                                         squeeze=False, **kwargs)

    def linsolve(self, *args, **kwargs) -> 'Structure':
        """
        Performs a linear elastostatic calculation with pre- and 
        post-processing.
        """
        return self.linear_static_analysis(*args, **kwargs)

    def linear_static_analysis(self, *args, **kwargs) -> 'Structure':
        """
        Performs a linear elastostatic calculation with pre- and 
        post-processing.
        """
        self._preproc_linstat_(*args,  **kwargs)
        self._proc_linstat_(*args, **kwargs)
        return self._postproc_linstat_(*args, **kwargs)

    def free_vibration_analysis(self, *args, **kwargs) -> 'Structure':
        """
        Performs a linear elastostatic calculation with pre- and 
        post-processing.
        """
        self._preproc_free_vib_(*args,  **kwargs)
        self._proc_free_vib_(*args, **kwargs)
        return self._postproc_free_vib_(*args, **kwargs)

    def effective_modal_masses(self, actions: ndarray,
                               modes: ndarray) -> ndarray:
        """
        Returns effective modal masses of several modes of vibration. 
        The call forwards all parameters to 
        :func:`sigmaepslion.solid.fem.dyn.effective_modal_masses`, see the 
        docs there for the details.

        Parameters
        ----------
        actions : numpy.ndarray
            One or more nodal displacement vectors representing rigid
            body motions.
        modes : numpy.ndarray
            Modes of vibration.        

        Returns
        -------
        numpy.ndarray
            An array of effective mass values.

        See also
        --------
        :func:`sigmaepslion.solid.fem.dyn.effective_modal_masses`

        """
        M_sparse = self.mesh.mass_matrix(penalize=False)
        return effective_modal_masses(M_sparse, actions, modes)

    def Rayleigh_quotient(self, u: ndarray, f: ndarray) -> ndarray:
        """
        Returns Rayleigh's quotient. The call forwards all parameters
        to :func:`sigmaepslion.solid.fem.dyn.Rayleigh_quotient`, see the 
        docs there for the details. If there are no actions specified, 
        the function feeds it with the results from a linear elastic solution.

        Parameters
        ----------
        u : ndarray
            One or more nodal displacement vectors.
        f : ndarray
            One or more nodal load vectors.

        Returns
        -------
        float or numpy.ndarray
            One or more floats, depending on the input.

        See also
        --------
        :func:`sigmaepslion.solid.fem.dyn.Rayleigh_quotient`

        """
        M_sparse = self.mass_matrix(penalize=True)
        return Rayleigh_quotient(M_sparse, u, f=f)

    def _initialize_(self, *args, **kwargs) -> 'Structure':
        """
        Initializes the structure. This includes data initialization 
        for the cells and happens during the preprocessor. Returns the object 
        for continuation.

        See also
        --------
        :func:`preprocess`

        """
        blocks = self.mesh.cellblocks(inclusive=True)
        for block in blocks:
            nE = len(block.celldata)
            # populate material stiffness matrices
            if not 'mat' in block.celldata.fields:
                C = block.material_stiffness_matrix()
                if not len(C.shape) == 3:
                    C = repeat(C, nE)
                block.celldata._wrapped['mat'] = C
            # populate frames
            if not 'frames' in block.celldata.fields:
                frames = repeat(block.frame.show(), nE)
                block.celldata._wrapped['frames'] = frames
        return self

    def _assemble_linstat_(self, *args, **kwargs):
        mesh = self._wrapped
        gnum = mesh.element_dof_numbering()
        # get raw data
        mesh.nodal_load_vector(squeeze=False)
        mesh.cell_load_vector(assemble=False, transform=False, squeeze=False)
        mesh.elastic_stiffness_matrix(sparse=False, transform=False)
        # condensate
        mesh.condensate_cell_fixity()
        # get condensated data
        f_nodal = mesh.nodal_load_vector(squeeze=False)
        f_bulk = mesh.cell_load_vector(assemble=False, transform=False,
                                       squeeze=False)
        K_bulk = mesh.elastic_stiffness_matrix(sparse=False, transform=False)
        # transform to global
        dcm = mesh.direction_cosine_matrix(target='global')
        K_bulk = tr_element_matrices_bulk(K_bulk, dcm)
        f_bulk = tr_nodal_loads_bulk(f_bulk, dcm)
        # assemble nodal load vector, the stiffness matrix gets assembled
        # in the solver
        nX = len(mesh.pd) * mesh.NDOFN
        f_bulk = assemble_load_vector(f_bulk, gnum, nX)
        f = f_nodal + f_bulk
        # penalize essential boundary conditions
        Kp = mesh.essential_penalty_matrix()
        fp = np.zeros(f.shape[0], dtype=float)
        for c in self.constraints:
            Kpc, fpc = c.assemble(mesh)
            Kp += Kpc
            fp += fpc
        if len(f.shape) == 1:
            f[:] += fp
        else:
            for i in range(f.shape[1]):
                f[:, i] += fp
        # create solver
        self._static_solver_ = StaticSolver(K_bulk, Kp.tocoo(), f, gnum,
                                            regular=False, mesh=mesh)
        return self

    def _assemble_free_vib_(self, *args, **kwargs):
        mesh = self._wrapped
        gnum = mesh.element_dof_numbering()
        # get raw data
        mesh.nodal_load_vector(squeeze=False)
        mesh.cell_load_vector(assemble=False, transform=False, squeeze=False)
        mesh.elastic_stiffness_matrix(sparse=False, transform=False)
        mesh.consistent_mass_matrix(sparse=False, transform=False)
        # condensate
        mesh.condensate_cell_fixity()
        # get condensated data
        M_sparse = mesh.mass_matrix()
        K_bulk = mesh.elastic_stiffness_matrix(sparse=False, transform=False)
        # transform to global
        dcm = mesh.direction_cosine_matrix(target='global')
        K_bulk = tr_element_matrices_bulk(K_bulk, dcm)
        # penalize essential boundary conditions
        Kp = mesh.essential_penalty_matrix()
        for c in self.constraints:
            Kpc, _ = c.assemble(mesh)
            Kp += Kpc
        # create solver
        self._dynamic_solver_ = DynamicSolver(K_bulk, Kp.tocoo(), M_sparse, gnum,
                                              regular=False, mesh=mesh)
        return self

    def _preproc_linstat_(self, *args, **kwargs) -> 'Structure':
        self._initialize_(*args, **kwargs)
        self._assemble_linstat_(*args, **kwargs)
        return self

    def _preproc_free_vib_(self, *args, **kwargs) -> 'Structure':
        self._initialize_(*args, **kwargs)
        self._assemble_free_vib_(*args, **kwargs)
        return self

    def _proc_linstat_(self, *args, **kwargs) -> 'Structure':
        self._static_solver_.solve(*args, **kwargs)
        return self

    def _proc_free_vib_(self, *args, **kwargs) -> 'Structure':
        self._dynamic_solver_.natural_circular_frequencies(*args, **kwargs)
        return self

    def _postproc_linstat_(self, *args, cleanup: bool = False,
                           **kwargs) -> 'Structure':
        solver = self._static_solver_
        mesh = self.mesh
        nDOFN = mesh.NDOFN
        nN = len(mesh.pd)
        # dof solution
        u = solver.u
        nRHS = 1 if len(u.shape) == 1 else u.shape[-1]
        if nRHS == 1:
            u = np.reshape(u, (nN, nDOFN))
        else:
            u = np.reshape(u, (nN, nDOFN, nRHS))
        self.mesh.pd.dofsol = u
        # reaction forces
        r = solver.r
        if nRHS == 1:
            r = np.reshape(r, (nN, nDOFN))
        else:
            r = np.reshape(r, (nN, nDOFN, nRHS))
        self.mesh.pd.reactions = r
        # nodal loads
        f = solver.f
        if nRHS == 1:
            f = np.reshape(f, (nN, nDOFN))
        else:
            f = np.reshape(f, (nN, nDOFN, nRHS))
        self.mesh.pd.forces = f
        # postproc block results
        self.mesh.postprocess(*args, **kwargs)
        if cleanup:
            self.cleanup()
        return self

    def _postproc_free_vib_(self, *, cleanup: bool = False,
                            **kwargs) -> 'Structure':
        mesh = self.mesh
        solver = self._dynamic_solver_
        self._natural_circular_frequencies = solver.frequencies
        vshapes = solver.modal_shapes
        nN = len(mesh.pd)
        nDOF = mesh.__class__.NDOFN
        nS = vshapes.shape[-1]
        mesh.pd.vshapes = vshapes.reshape(nN, nDOF, nS)
        if cleanup:
            self.cleanup()
        return self

    def cleanup(self) -> 'Structure':
        """
        Destroys the solver and returns the object for continuation.

        Returns
        -------
        Structure

        """
        self._static_solver_ = None
        self._dynamic_solver_ = None
        return self



# -*- coding: utf-8 -*-
import numpy as np
from numba import njit

from neumann.numint import GaussPoints as Gauss

from polymesh.tri.triutils import area_tri_bulk
from polymesh.utils import cells_coords
from polymesh.cells import Q4 as Quadrilateral

from ..model.membrane import Membrane
from ..model.plate import Plate

from .elem import FiniteElement
from .metaelem import ABCFiniteElement as ABC

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def area_Q4_bulk(ecoords : np.ndarray):
    nE = len(ecoords)
    res = np.zeros(nE, dtype=ecoords.dtype)
    res += area_tri_bulk(ecoords[:, :3, :])
    res += area_tri_bulk(ecoords[:, np.array([0, 2, 3]), :])
    return res


class Q4M(ABC, Membrane, Quadrilateral, FiniteElement):
    
    qrule = 'full'
    quadrature = {
        'full' : Gauss(2, 2), 
        'selective' : {
            (0, 1) : 'full',
            (2,) : 'reduced'
            },
        'reduced' : Gauss(1, 1)
        }
        
    def areas(self, *args, coords=None, topo=None, **kwargs):
        """This shadows the original geometrical implementation."""
        if coords is None:
            coords = self.container.root().coords()
        topo = self.topology().to_numpy() if topo is None else topo
        return area_Q4_bulk(cells_coords(coords, topo))
       

class Q4P(ABC, Plate, Quadrilateral, FiniteElement):
    
    qrule = 'selective'
    quadrature = {
        'full' : Gauss(2, 2), 
        'selective' : {
            (0, 1, 2) : 'full',
            (3, 4) : 'reduced'
            },
        'reduced' : Gauss(1, 1)
        }
        
    def areas(self, *args, coords=None, topo=None, **kwargs):
        """This shadows the original geometrical implementation."""
        if coords is None:
            coords = self.container.root().coords()
        topo = self.topology().to_numpy() if topo is None else topo
        self.local_coordinates(topo=topo)
        return area_Q4_bulk(cells_coords(coords, topo))
    
    
if __name__ == '__main__':
    from sympy import symbols

    r, s = symbols('r s')

    N1 = 0.125*(1-r)*(1-s)
    N2 = 0.125*(1+r)*(1-s)
    N3 = 0.125*(1+r)*(1+s)
    N4 = 0.125*(1-r)*(1+s)



from dewloosh.core.downloads import download_stand, _download_file
from polymesh import PolyData
from typing import Union



def download_bernoulli_console_json_linstat():  # pragma: no cover
    """
    Downloads the description of a simple bernoulli console as a json file.
   
    Returns
    -------
    str
        A path to a file on your filesystem.

    Example
    --------
    >>> from sigmaepsilon.examples import download_bernoulli_console_json_20B2
    >>> jsonpath = download_bernoulli_console_json_20B2()
    
    """
    return _download_file('console_bernoulli_linstat.json')[0]


"""def stand_vtk(read=False) -> Union[str, PolyData]:
    vtkpath = download_stand()
    if read:
        return PolyData.read(vtkpath)
    else:
        return vtkpath"""


# -*- coding: utf-8 -*-
from abc import abstractmethod

from neumann.array import atleast2d, atleast3d

from .utils import (model_strains, model_strains_multi,
                    stresses_from_strains, stresses_from_strains_multi)
from ..utils import topo_to_gnum
from ..cells.metaelem import FemModel


__all__ = ['Solid']


class Solid(FemModel):
    
    dofs = ('UX', 'UY', 'UZ')
    strn = ('exx', 'eyy', 'ezz', 'eyz', 'exz', 'exy')

    NDOFN = 3
    NSTRE = 6

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def material_stiffness_matrix(self, *args, **kwargs):
        return self._wrapped.mat.to_numpy()

    def model_stiffness_matrix(self, *args, **kwargs):
        return self.material_stiffness_matrix()

    @classmethod
    def HMH(cls, *args,  **kwargs):
        raise NotImplementedError

    @classmethod
    def model_strains(cls, dofsol1d, gnum, B, *args, **kwargs):
        if len(dofsol1d.shape) == 2:
            return model_strains_multi(dofsol1d, gnum, B)
        return model_strains(dofsol1d, gnum, B)

    def strains_at(self, lcoords, *args,  z=None, topo=None, **kwargs):
        topo = self.topology().to_numpy() if topo is None else topo
        gnum = topo_to_gnum(topo, self.NDOFN)
        lcoords = atleast2d(lcoords)
        dshp = self.shape_function_derivatives(lcoords)
        ecoords = self.local_coordinates(topo=topo)
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ecoords)
        B = self.strain_displacement_matrix(dshp=dshp, jac=jac)   
        dofsol = self.pointdata.dofsol.to_numpy()
        dofsol = atleast3d(dofsol, back=True)  
        return self.model_strains(dofsol, gnum, B)

    def stresses_at(self, *args, z=None, topo=None, **kwargs):
        """
        Returns stresses for every node of every element.
        """
        # The next code line returns either generalized strains
        # or material strains, depending on the value of z
        strns = self.strains_at(*args, z=z, topo=topo, **kwargs)
        C = self.model_stiffness_matrix()
        if 'HMH' in args:
            return self.HMH(stresses_from_strains(C, strns))
        return stresses_from_strains(C, strns)

    def stresses_at_nodes(self, *args, **kwargs):
        """
        Returns stresses for every node of every element.
        """
        return self.stresses_at(self.lcoords(), *args, **kwargs)

    def stresses_at_centers(self, *args, **kwargs):
        """
        Returns stresses at the centre of every element.
        """
        return self.stresses_at(self.lcenter(), *args, **kwargs)



# -*- coding: utf-8 -*-
from typing import Callable

import numpy as np
from numpy import ndarray

from neumann.array import atleastnd
from polymesh.pointdata import PointData as MeshPointData


class PointData(MeshPointData):
    """
    A subclass of :class:`polymesh.pointdata.PointData` to handle data related 
    to the pointcloud of a finite element mesh.

    Parameters
    ----------
    fixity : numpy.ndarray, Optional
        A 2d boolean or float array describing essential boundary conditions.
        Default is None.
        
    loads : numpy.ndarray, Optional
        A 2d or 3d float array of nodal loads.
        Default is None.
        
    mass : numpy.ndarray, Optional
        2d array of nodal masses for each degree of freedom of a node.
        Default is None.
        
    fields : dict, Optional
        Every value of this dictionary is added to the dataset. Default is `None`.
        
    **kwargs : dict, Optional
        For every key and value pair where the value is a numpy array
        with a matching shape (has entries for all points), the key
        is considered as a field and the value is added to the database.
        
    Examples
    --------
    We will define the point-related data of a simple linear console. At the
    moment this means the specification of cordinates, nodal loads, and boundary 
    conditions. The common properties of these attributes is that each of these
    can be best described by arrays having the same length as the pointcloud itself.
    Import the necessary stuff:
    
    >>> from sigmaepsilon.solid import Structure, LineMesh, PointData
    >>> from neumann.linalg import linspace, Vector
    >>> from polymesh.space import StandardFrame, PointCloud, frames_of_lines
    >>> import numpy as np
    
    Define a coordinate frame, and a coordinate array,

    >>> GlobalFrame = StandardFrame(dim=3)
    >>> nElem = 20  # number of finite elements to use
    >>> p0 = np.array([0., 0., 0.])
    >>> p1 = np.array([L, 0., 0.])
    >>> coords = linspace(p0, p1, nElem+1)
    >>> coords = PointCloud(coords, frame=GlobalFrame).show()

    two numpy arrays resembling the boundary conditions,

    >>> # support at the leftmost, load at the rightmost node
    >>> loads = np.zeros((coords.shape[0], 6))
    >>> fixity = np.zeros((coords.shape[0], 6)).astype(bool)
    >>> global_load_vector = Vector([0., 0, F], frame=GlobalFrame).show()
    >>> loads[-1, :3] = global_load_vector
    >>> fixity[0, :] = True

    and finally we can define our dataset.  

    >>> pd = PointData(coords=coords, frame=GlobalFrame, 
    >>>                loads=loads, fixity=fixity)

    See also
    --------
    :class:`sigmaepsilon.solid.fem.cells.celldata.CellData`
    :class:`polymesh.PolyData`
    :class:`polymesh.CellData`
    :class:`awkward.Record`
    
    """

    _attr_map_ = {
        'loads': 'loads',
        'mass': 'mass',
        'fixity': 'fixity',
        'dofsol' : 'dofsol',
        'shapes' : '_shapes_',
        'reactions' : 'reactions',
        'forces' : 'forces',
    }
    NDOFN = 6

    def __init__(self, *args, fixity:ndarray=None, loads:ndarray=None, 
                 mass:ndarray=None, fields:dict=None, **kwargs):
        amap = self.__class__._attr_map_

        if fields is not None:
            fixity = fields.pop(amap['fixity'], fixity)
            loads = fields.pop(amap['loads'], loads)
            mass = fields.pop(amap['mass'], mass)

        super().__init__(*args, fields=fields, **kwargs)

        if self.db is not None:
            nP = len(self)
            NDOFN = self.__class__.NDOFN

            if fixity is None:
                fixity = np.zeros((nP, NDOFN), dtype=bool)
            else:
                assert isinstance(fixity, np.ndarray) and len(
                    fixity.shape) == 2
            self.fixity = fixity

            if loads is None:
                loads = np.zeros((nP, NDOFN, 1))
            if loads is not None:
                assert isinstance(loads, np.ndarray)
                if loads.shape[0] == nP:
                    self.loads = loads
                elif loads.shape[0] == nP * NDOFN:
                    loads = atleastnd(loads, 2, back=True)
                    self.loads = loads.reshape(nP, NDOFN, loads.shape[-1])

            if mass is not None:
                if isinstance(mass, float) or isinstance(mass, int):
                    raise NotImplementedError
                elif isinstance(mass, np.ndarray):
                    self.mass = mass

    @property
    def _dbkey_fixity_(self) -> str:
        return self.__class__._attr_map_['fixity']
    
    @property
    def has_fixity(self):
        return self._dbkey_fixity_ in self._wrapped.fields
    
    @property
    def fixity(self) -> ndarray:
        """Returns fixity information as a 2d numpy array."""
        return self._wrapped[self.__class__._attr_map_['fixity']].to_numpy()

    @fixity.setter
    def fixity(self, value: ndarray):
        """Sets essential boundary conditions."""
        assert isinstance(value, ndarray)
        self._wrapped[self.__class__._attr_map_['fixity']] = value

    @property
    def loads(self) -> ndarray:
        """Returns the nodal load vector."""
        return self._wrapped[self.__class__._attr_map_['loads']].to_numpy()

    @loads.setter
    def loads(self, value: ndarray):
        """Sets the nodal load vector."""
        assert isinstance(value, ndarray)
        self._wrapped[self.__class__._attr_map_['loads']] = value
        
    @property
    def mass(self) -> ndarray:
        """Retruns nodal masses."""
        return self._wrapped[self.__class__._attr_map_['mass']].to_numpy()

    @mass.setter
    def mass(self, value: ndarray):
        """Sets nodal masses."""
        assert isinstance(value, ndarray)
        self._wrapped[self.__class__._attr_map_['mass']] = value
        
    @property
    def dofsol(self) -> ndarray:
        """Returns the displacement vector."""
        return self._wrapped[self.__class__._attr_map_['dofsol']].to_numpy()

    @dofsol.setter
    def dofsol(self, value: ndarray):
        assert isinstance(value, ndarray)
        self._wrapped[self.__class__._attr_map_['dofsol']] = value
        
    @property
    def vshapes(self) -> ndarray:
        """Returns the modal shapes."""
        return self._wrapped[self.__class__._attr_map_['shapes']].to_numpy()

    @vshapes.setter
    def vshapes(self, value: ndarray):
        assert isinstance(value, ndarray)
        self._wrapped[self.__class__._attr_map_['shapes']] = value  
        
    @property
    def reactions(self) -> ndarray:
        """Returns the vector of reaction forces."""
        return self._wrapped[self.__class__._attr_map_['reactions']].to_numpy()

    @reactions.setter
    def reactions(self, value: ndarray):
        assert isinstance(value, ndarray)
        self._wrapped[self.__class__._attr_map_['reactions']] = value   
        
    @property
    def forces(self) -> ndarray:
        return self._wrapped[self.__class__._attr_map_['forces']].to_numpy()

    @forces.setter
    def forces(self, value: ndarray):
        assert isinstance(value, ndarray)
        self._wrapped[self.__class__._attr_map_['forces']] = value    


def flatten_pd(default=True):
    def decorator(fnc: Callable):
        def inner(*args, flatten: bool = default, **kwargs):
            if flatten:
                x = fnc(*args, **kwargs)
                if len(x.shape) == 2:
                    return x.flatten()
                else:
                    nN, nDOFN, nRHS = x.shape
                    return x.reshape((nN * nDOFN, nRHS))
            else:
                return fnc(*args, **kwargs)
        inner.__doc__ = fnc.__doc__
        return inner
    return decorator


# -*- coding: utf-8 -*-
from neumann.numint import GaussPoints as Gauss

from polymesh.cells import Q9 as Quadrilateral

from ..model.membrane import Membrane
from ..model.plate import Plate

from .elem import FiniteElement
from .metaelem import ABCFiniteElement as ABC

class Q9M(ABC, Membrane, Quadrilateral, FiniteElement):

    qrule = 'full'
    quadrature = {
        'full': Gauss(3, 3),
        'selective': {
            (0, 1): 'full',
            (2,): 'reduced'
        },
        'reduced': Gauss(2, 2)
    }


class Q9P(ABC, Plate, Quadrilateral, FiniteElement):

    qrule = 'selective'
    quadrature = {
        'full': Gauss(3, 3),
        'selective': {
            (0, 1, 2): 'full',
            (3, 4): 'reduced'
        },
        'reduced': Gauss(2, 2)
    }



# -*- coding: utf-8 -*-
from typing import Iterable
import numpy as np
from numpy import ndarray
from abc import abstractmethod

from linkeddeepdict import LinkedDeepDict
from linkeddeepdict.tools.kwargtools import getasany, allinkwargs, anyinkwargs
from linkeddeepdict.tools.dtk import parsedicts_addr


class MetaSurface(LinkedDeepDict):
    """
    Base object implementing methods that both a folder (a shell) and a
    file (a layer) can posess.
    """
    
    __layerclass__ = None
    
    @property
    def angle(self):
        """Returns the angle of a layer or a laminate."""
        _angle = self.get('angle', None)
        if _angle is None:
            return None if self.is_root() else self.parent.angle
        else:
            return _angle

    @angle.setter
    def angle(self, value):
        """Sets the angle of a layer or a laminate."""
        if self.__layerclass__ is None:
            self['angle'] = value
        else:
            for layer in self.containers(dtype=self.__layerclass__):
                layer['angle'] = value
            self['angle'] = 0
            del self['angle']
            
    @property
    def hooke(self):
        """Returns the Hooke model of a layer or a laminate."""
        _hooke = self.get('hooke', None)
        if _hooke is None:
            return None if self.is_root() else self.parent.hooke
        else:
            return _hooke

    @hooke.setter
    def hooke(self, value):
        """Sets the Hooke model of a layer or a laminate."""
        if self.__layerclass__ is None:
            self['hooke'] = value
        else:
            for layer in self.containers(dtype=self.__layerclass__):
                layer['hooke'] = value
            self['hooke'] = 0
            del self['hooke']
    
    @property
    def t(self):
        """Returns the thickness of a layer."""
        return self.get('thickness', None)
    
    @t.setter
    def t(self, value):
        """Sets the thickness of a layer."""
        if self.__layerclass__ is None:
            self['thickness'] = value
        else:
            raise RuntimeError
        

class Layer(MetaSurface):
    """
    Helper base class for layers of a laminate.
    """

    __loc__ = [-1., 0., 1.]
    __shape__ = (8, 8)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hooke = getasany(['material', 'm', 'hooke'], **kwargs)
        # set thickness
        self.tmin = None
        self.tmax = None
        self.t = None
        if allinkwargs(['tmin', 'tmax'], **kwargs):
            self.tmin = kwargs.get('tmin', None)
            self.tmax = kwargs.get('tmax', None)
            self.t = self.tmax-self.tmin
        elif anyinkwargs(['t', 'thickness'], **kwargs):
            self.t = getasany(['t', 'thickness'], **kwargs)
            if 'tmin' in kwargs:
                self.tmin = kwargs['tmin']
                self.tmax = self.tmin + self.t
            elif 'tmax' in kwargs:
                self.tmax = kwargs['tmax']
                self.tmin = self.tmax - self.t
            else:
                self.tmin = (-1) * self.t / 2
                self.tmax = self.t / 2
                
    def loc_to_z(self, loc) -> float:
        """
        Returns height of a local point by linear interpolation.
        Local coordinate is expected between -1 and 1.
        """
        return 0.5 * ((self.tmax + self.tmin) + loc * (self.tmax - self.tmin))

    @abstractmethod
    def stiffness_matrix(self):
        raise NotImplementedError
    
    
class Surface(MetaSurface):
    """
    Helper base class for laminates.
    """

    __layerclass__ = Layer

    def Layer(self, *args, **kwargs):
        """
        Returns a Layer compatible with the model.
        """
        return self.__layerclass__(*args, **kwargs)
    
    def Hooke(self):
        raise NotImplementedError

    def layers(self) -> Iterable:
        """
        Returns the layers of the laminate.
        """
        return [layer for layer in self.containers(dtype=self.__layerclass__)]

    def iterlayers(self):
        """
        Returns the layers of the laminate as a generator.
        """
        return self.containers(dtype=self.__layerclass__)

    def stiffness_matrix(self) -> np.ndarray:
        """
        Assembles and returns the stiffness matrix.
        """
        self._set_layers()
        res = np.zeros(self.__layerclass__.__shape__)
        for layer in self.iterlayers():
            res += layer.stiffness_matrix()
        return res

    def _set_layers(self):
        """
        Sets thickness ranges for the layers.
        """
        layers = self.layers()
        t = sum([layer.t for layer in layers])
        layers[0].tmin = -t/2
        nLayers = len(layers)
        for i in range(nLayers-1):
            layers[i].tmax = layers[i].tmin + layers[i].t
            layers[i+1].tmin = layers[i].tmax
        layers[-1].tmax = t/2
        for layer in layers:
            layer.zi = [layer.loc_to_z(l_) for l_ in layer.__loc__]
        return True
    
    @classmethod
    def from_dict(cls, d: dict = None, **kwargs) -> 'Surface':  
        res = cls(**d)
        for addr, value in parsedicts_addr(d, inclusive=True):
            if len(addr) == 0:
                continue
            if 'hooke' in value:
                subcls = cls.__layerclass__
            else: 
                continue
            value['key'] = addr[-1]
            res[addr] = subcls(**value)
        return res



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def calc_beam_stresses_2d(forces: ndarray, factors: ndarray):
    """
    Returns the stresses in the order s11, s22, s33, s12, s13, s23.
    
    Parameters
    ----------
    forces : numpy.ndarray
        A 2d float array of shape (nX, nSTRE) where 'nX' is the number of records
        in the data and 'nSTRE' is the number of internal force components.
        
    factors (nS, nSTRE, nPS)
        Factors for every stress (nS) and internal force (nSTRE) component
        at a nPS number of points of a section.
    
    Returns
    -------
    numpy.ndarray
        A 3d array of stresses of shape (nX, nPS, nS).
        
    """
    nX, nSTRE = forces.shape
    nS, _, nPS = factors.shape
    res = np.zeros((nX, nPS, nS))
    for iX in prange(nX):
        for iPS in prange(nPS):
            for iS in prange(nS):
                for iSTRE in range(nSTRE):
                    res[iX, iPS, iS] += \
                        forces[iX, iSTRE] * factors[iS, iSTRE, iPS]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def calc_beam_stresses_4d(forces: ndarray, factors: ndarray):
    """
    Returns the stresses in the order s11, s22, s33, s12, s13, s23.
    
    Parameters
    ----------
    forces : numpy.ndarray
        A 4d float array of shape (nE, nP, nSTRE, nRHS), representing
        all internal forces (nSTRE) for multiple elements(nE), load cases(nRHS) 
        and evaluation points (nP).
        
    factors (nS, nSTRE, nPS)
        Factors for every stress (nS) and internal force (nSTRE) component
        at a nPS number of points in the section.
    
    Returns
    -------
    numpy.ndarray
        A 5d array of stresses of shape (nE, nP, nPS, nS, nRHS).
        
    """
    nE, nP, nSTRE, nRHS = forces.shape
    nS, _, nPS = factors.shape
    res = np.zeros((nE, nP, nPS, nS, nRHS))
    for iE in prange(nE):
        for iP in prange(nP):
            for iPS in prange(nPS):
                for iS in prange(nS):
                    for iRHS in prange(nRHS):
                        for iSTRE in range(nSTRE):
                            res[iE, iP, iPS, iS, iRHS] += \
                                forces[iE, iP, iSTRE, iRHS] * \
                                factors[iS, iSTRE, iPS]
    return res
